"""Buffers to store NumPy arrays and transitions in."""

from typing import Any, Mapping, Optional, Tuple

import numpy as np
from stable_baselines3.common import vec_env

from imitation.data import types


def num_samples(data: Mapping[Any, np.ndarray]) -> int:
    """Computes the number of samples contained in `data`.

    Args:
        data: A Mapping from keys to NumPy arrays.

    Returns:
        The unique length of the first dimension of arrays contained in `data`.

    Raises:
        ValueError: The length is not unique.
    """
    n_samples_list = [arr.shape[0] for arr in data.values()]
    n_samples_np = np.unique(n_samples_list)
    if len(n_samples_np) > 1:
        raise ValueError("Keys map to different length values.")
    return int(n_samples_np[0])


class Buffer:
    """A FIFO ring buffer for NumPy arrays of a fixed shape and dtype.

    Supports random sampling with replacement.
    """

    capacity: int
    """The number of data samples that can be stored in this buffer."""

    sample_shapes: Mapping[str, Tuple[int, ...]]
    """The shapes of each data sample stored in this buffer."""

    _arrays: Mapping[str, np.ndarray]
    """The underlying NumPy arrays (which actually store the data)."""

    _n_data: int
    """The number of samples currently stored in this buffer.

    An integer in `range(0, self.capacity + 1)`. This attribute is the return
    value of `self.size()`.
    """

    _idx: int
    """The index of the first row that new data should be written to.

    An integer in `range(0, self.capacity)`.
    """

    def __init__(
        self,
        capacity: int,
        sample_shapes: Mapping[str, Tuple[int, ...]],
        dtypes: Mapping[str, np.dtype],
    ):
        """Constructs a Buffer.

        Args:
            capacity: The number of samples that can be stored.
            sample_shapes: A dictionary mapping string keys to the shape of
                samples associated with that key.
            dtypes (`np.dtype`-like): A dictionary mapping string keys to the dtype
                of samples associated with that key.

        Raises:
            KeyError: `sample_shapes` and `dtypes` have different keys.
        """
        if sample_shapes.keys() != dtypes.keys():
            raise KeyError("sample_shape and dtypes keys don't match")
        self.capacity = capacity
        self.sample_shapes = {k: tuple(shape) for k, shape in sample_shapes.items()}
        self._arrays = {
            k: np.zeros((capacity,) + shape, dtype=dtypes[k])
            for k, shape in self.sample_shapes.items()
        }
        self._n_data = 0
        self._idx = 0

    @classmethod
    def from_data(
        cls,
        data: Mapping[str, np.ndarray],
        capacity: Optional[int] = None,
        truncate_ok: bool = False,
    ) -> "Buffer":
        """Constructs and return a Buffer containing the provided data.

        Shapes and dtypes are automatically inferred.

        Args:
            data: A dictionary mapping keys to data arrays. The arrays may differ
                in their shape, but should agree in the first axis.
            capacity: The Buffer capacity. If not provided, then this is automatically
                set to the size of the data, so that the returned Buffer is at full
                capacity.
            truncate_ok: Whether to error if `capacity` < the number of samples in
                `data`. If False, then only store the last `capacity` samples from
                `data` when overcapacity.

        Examples:
            In the follow examples, suppose the arrays in `data` are length-1000.

            `Buffer` with same capacity as arrays in `data`::

                Buffer.from_data(data)

            `Buffer` with larger capacity than arrays in `data`::

                Buffer.from_data(data, 10000)

            `Buffer with smaller capacity than arrays in `data`. Without
            `truncate_ok=True`, `from_data` will error::

                Buffer.from_data(data, 5, truncate_ok=True)

        Returns:
            Buffer of specified `capacity` containing provided `data`.

        Raises:
            ValueError: `data` is empty.
            ValueError: `data` has items mapping to arrays differing in the
                length of their first axis.
        """
        data_capacities_list = [arr.shape[0] for arr in data.values()]
        data_capacities = np.unique(data_capacities_list)
        if len(data) == 0:
            raise ValueError("No keys in data.")
        if len(data_capacities) > 1:
            raise ValueError("Keys map to different length values")
        if capacity is None:
            capacity = data_capacities[0]

        sample_shapes = {k: arr.shape[1:] for k, arr in data.items()}
        dtypes = {k: arr.dtype for k, arr in data.items()}
        buf = cls(capacity, sample_shapes, dtypes)
        buf.store(data, truncate_ok=truncate_ok)
        return buf

    def store(self, data: Mapping[str, np.ndarray], truncate_ok: bool = False) -> None:
        """Stores new data samples, replacing old samples with FIFO priority.

        Args:
            data: A dictionary mapping keys `k` to arrays with shape
                `(n_samples,) + self.sample_shapes[k]`, where `n_samples` is less
                than or equal to `self.capacity`.
            truncate_ok: If False, then error if the length of `transitions` is
                greater than `self.capacity`. Otherwise, store only the final
                `self.capacity` transitions.

        Raises:
            ValueError: `data` is empty.
            ValueError: If `n_samples` is greater than `self.capacity`.
            ValueError: data is the wrong shape.
        """
        expected_keys = set(self.sample_shapes.keys())
        missing_keys = expected_keys.difference(data.keys())
        unexpected_keys = set(data.keys()).difference(expected_keys)
        if len(missing_keys) > 0:
            raise ValueError(f"Missing keys {missing_keys}")
        if len(unexpected_keys) > 0:
            raise ValueError(f"Unexpected keys {unexpected_keys}")

        n_samples = num_samples(data)
        if n_samples == 0:
            raise ValueError("Trying to store empty data.")
        if n_samples > self.capacity:
            if not truncate_ok:
                raise ValueError("Not enough capacity to store data.")
            else:
                data = {k: arr[-self.capacity :] for k, arr in data.items()}

        for k, arr in data.items():
            if arr.shape[1:] != self.sample_shapes[k]:
                raise ValueError(f"Wrong data shape for {k}")

        new_idx = self._idx + n_samples
        if new_idx > self.capacity:
            n_remain = self.capacity - self._idx
            # Need to loop around the buffer. Break into two "easy" calls.
            self._store_easy({k: arr[:n_remain] for k, arr in data.items()})
            assert self._idx == 0
            self._store_easy({k: arr[n_remain:] for k, arr in data.items()})
        else:
            self._store_easy(data)

    def _store_easy(self, data: Mapping[str, np.ndarray]) -> None:
        """Stores new data samples, replacing old samples with FIFO priority.

        Requires that `size(data) <= self.capacity - self._idx`, where `size(data)` is
        the number of rows in every array in `data.values()`. Updates `self._idx`
        to be the insertion point of the next call to `_store_easy` call,
        looping back to `self._idx = 0` if necessary.

        Also updates `self._n_data`.

        Args:
            data: Same as in `self.store`'s docstring, except with the additional
                constraint `size(data) <= self.capacity - self._idx`.
        """
        n_samples = num_samples(data)
        assert n_samples <= self.capacity - self._idx
        idx_hi = self._idx + n_samples
        for k, arr in data.items():
            self._arrays[k][self._idx : idx_hi] = arr
        self._idx = idx_hi % self.capacity
        self._n_data = min(self._n_data + n_samples, self.capacity)

    def sample(self, n_samples: int) -> Mapping[str, np.ndarray]:
        """Uniformly sample `n_samples` samples from the buffer with replacement.

        Args:
            n_samples: The number of samples to randomly sample.

        Returns:
            samples (np.ndarray): An array with shape
                `(n_samples) + self.sample_shape`.

        Raises:
            ValueError: The buffer is empty.
        """
        if self.size() == 0:
            raise ValueError("Buffer is empty")
        ind = np.random.randint(self.size(), size=n_samples)
        return {k: buffer[ind] for k, buffer in self._arrays.items()}

    def size(self) -> int:
        """Returns the number of samples stored in the buffer."""
        assert 0 <= self._n_data <= self.capacity
        return self._n_data


class ReplayBuffer:
    """Buffer for Transitions."""

    capacity: int
    """The number of data samples that can be stored in this buffer."""

    def __init__(
        self,
        capacity: int,
        venv: Optional[vec_env.VecEnv] = None,
        *,
        obs_shape: Optional[Tuple[int, ...]] = None,
        act_shape: Optional[Tuple[int, ...]] = None,
        obs_dtype: Optional[np.dtype] = None,
        act_dtype: Optional[np.dtype] = None,
    ):
        """Constructs a ReplayBuffer.

        Args:
            capacity: The number of samples that can be stored.
            venv: The environment whose action and observation
                spaces can be used to determine the data shapes of the underlying
                buffers. Mutually exclusive with shape and dtype arguments.
            obs_shape: The shape of the observation space.
            act_shape: The shape of the action space.
            obs_dtype: The dtype of the observation space.
            act_dtype: The dtype of the action space.

        Raises:
            ValueError: Couldn't infer the observation and action shapes and dtypes
                from the arguments.
            ValueError: Specified both venv and shapes/dtypes.
        """
        params = (obs_shape, act_shape, obs_dtype, act_dtype)
        if venv is not None:
            if venv.observation_space.shape is not None:
                if obs_shape is not None:
                    raise ValueError(
                        "Cannot specify both observation shape and also environment "
                        "with an observation space that has a shape.",
                    )
                obs_shape = tuple(venv.observation_space.shape)

            if venv.observation_space.dtype is not None:
                if obs_dtype is not None:
                    raise ValueError(
                        "Cannot specify both observation dtype and also environment "
                        "with an observation space that has a dtype.",
                    )
                obs_dtype = venv.observation_space.dtype

            if venv.action_space.shape is not None:
                if act_shape is not None:
                    raise ValueError(
                        "Cannot specify both action shape and also environment "
                        "with an action space that has a shape.",
                    )
                act_shape = tuple(venv.action_space.shape)

            if venv.action_space.dtype is not None:
                if act_dtype is not None:
                    raise ValueError(
                        "Cannot specify both action dtype and also environment "
                        "with an action space that has a dtype.",
                    )
                act_dtype = venv.action_space.dtype
        else:
            if any(x is None for x in params):
                raise ValueError("Shape or dtype missing and no environment specified.")

        assert obs_shape is not None
        assert act_shape is not None
        assert obs_dtype is not None
        assert act_dtype is not None

        self.capacity = capacity
        sample_shapes = {
            "obs": obs_shape,
            "acts": act_shape,
            "next_obs": obs_shape,
            "dones": (),
            "infos": (),
        }
        dtypes = {
            "obs": obs_dtype,
            "acts": act_dtype,
            "next_obs": obs_dtype,
            "dones": np.dtype(bool),
            "infos": np.dtype(object),
        }
        self._buffer = Buffer(capacity, sample_shapes=sample_shapes, dtypes=dtypes)

    @classmethod
    def from_data(
        cls,
        transitions: types.Transitions,
        capacity: Optional[int] = None,
        truncate_ok: bool = False,
    ) -> "ReplayBuffer":
        """Construct and return a ReplayBuffer containing the provided data.

        Shapes and dtypes are automatically inferred, and the returned ReplayBuffer is
        ready for sampling.

        Args:
            transitions: Transitions to store.
            capacity: The ReplayBuffer capacity. If not provided, then this is
                automatically set to the size of the data, so that the returned Buffer
                is at full capacity.
            truncate_ok: Whether to error if `capacity` < the number of samples in
                `data`. If False, then only store the last `capacity` samples from
                `data` when overcapacity.

        Examples:
            `ReplayBuffer` with same capacity as arrays in `data`::

                ReplayBuffer.from_data(data)

            `ReplayBuffer` with larger capacity than arrays in `data`::

                ReplayBuffer.from_data(data, 10000)

            `ReplayBuffer with smaller capacity than arrays in `data`. Without
            `truncate_ok=True`, `from_data` will error::

                ReplayBuffer.from_data(data, 5, truncate_ok=True)

        Returns:
            A new ReplayBuffer.
        """
        obs = types.assert_not_dictobs(transitions.obs)
        obs_shape = obs.shape[1:]
        act_shape = transitions.acts.shape[1:]
        if capacity is None:
            capacity = obs.shape[0]
        instance = cls(
            capacity=capacity,
            obs_shape=obs_shape,
            act_shape=act_shape,
            obs_dtype=obs.dtype,
            act_dtype=transitions.acts.dtype,
        )
        instance.store(transitions, truncate_ok=truncate_ok)
        return instance

    def sample(self, n_samples: int) -> types.Transitions:
        """Sample obs-act-obs triples.

        Args:
            n_samples: The number of samples.

        Returns:
            A Transitions named tuple containing n_samples transitions.
        """
        sample = self._buffer.sample(n_samples)
        return types.Transitions(**sample)

    def store(self, transitions: types.Transitions, truncate_ok: bool = True) -> None:
        """Store obs-act-obs triples.

        Args:
            transitions: Transitions to store.
            truncate_ok: If False, then error if the length of `transitions` is
                greater than `self.capacity`. Otherwise, store only the final
                `self.capacity` transitions.

        Raises:
            ValueError: The arguments didn't have the same length.
        """  # noqa: DAR402
        trans_dict = types.dataclass_quick_asdict(transitions)
        # Remove unnecessary fields
        trans_dict = {k: trans_dict[k] for k in self._buffer.sample_shapes.keys()}
        self._buffer.store(trans_dict, truncate_ok=truncate_ok)

    def size(self) -> Optional[int]:
        """Returns the number of samples stored in the buffer."""
        return self._buffer.size()



"""Core code for adversarial imitation learning, shared between GAIL and AIRL."""
import abc
import dataclasses
import logging
from typing import Callable, Iterable, Iterator, Mapping, Optional, Type, overload

import numpy as np
import torch as th
import torch.utils.tensorboard as thboard
import tqdm
from stable_baselines3.common import (
    base_class,
    distributions,
    on_policy_algorithm,
    policies,
    vec_env,
)
from stable_baselines3.sac import policies as sac_policies
from torch.nn import functional as F

from imitation.algorithms import base
from imitation.data import buffer, rollout, types, wrappers
from imitation.rewards import reward_nets, reward_wrapper
from imitation.util import logger, networks, util


def compute_train_stats(
    disc_logits_expert_is_high: th.Tensor,
    labels_expert_is_one: th.Tensor,
    disc_loss: th.Tensor,
) -> Mapping[str, float]:
    """Train statistics for GAIL/AIRL discriminator.

    Args:
        disc_logits_expert_is_high: discriminator logits produced by
            `AdversarialTrainer.logits_expert_is_high`.
        labels_expert_is_one: integer labels describing whether logit was for an
            expert (0) or generator (1) sample.
        disc_loss: final discriminator loss.

    Returns:
        A mapping from statistic names to float values.
    """
    with th.no_grad():
        # Logits of the discriminator output; >0 for expert samples, <0 for generator.
        bin_is_generated_pred = disc_logits_expert_is_high < 0
        # Binary label, so 1 is for expert, 0 is for generator.
        bin_is_generated_true = labels_expert_is_one == 0
        bin_is_expert_true = th.logical_not(bin_is_generated_true)
        int_is_generated_pred = bin_is_generated_pred.long()
        int_is_generated_true = bin_is_generated_true.long()
        n_generated = float(th.sum(int_is_generated_true))
        n_labels = float(len(labels_expert_is_one))
        n_expert = n_labels - n_generated
        pct_expert = n_expert / float(n_labels) if n_labels > 0 else float("NaN")
        n_expert_pred = int(n_labels - th.sum(int_is_generated_pred))
        if n_labels > 0:
            pct_expert_pred = n_expert_pred / float(n_labels)
        else:
            pct_expert_pred = float("NaN")
        correct_vec = th.eq(bin_is_generated_pred, bin_is_generated_true)
        acc = th.mean(correct_vec.float())

        _n_pred_expert = th.sum(th.logical_and(bin_is_expert_true, correct_vec))
        if n_expert < 1:
            expert_acc = float("NaN")
        else:
            # float() is defensive, since we cannot divide Torch tensors by
            # Python ints
            expert_acc = _n_pred_expert.item() / float(n_expert)

        _n_pred_gen = th.sum(th.logical_and(bin_is_generated_true, correct_vec))
        _n_gen_or_1 = max(1, n_generated)
        generated_acc = _n_pred_gen / float(_n_gen_or_1)

        label_dist = th.distributions.Bernoulli(logits=disc_logits_expert_is_high)
        entropy = th.mean(label_dist.entropy())

    return {
        "disc_loss": float(th.mean(disc_loss)),
        "disc_acc": float(acc),
        "disc_acc_expert": float(expert_acc),  # accuracy on just expert examples
        "disc_acc_gen": float(generated_acc),  # accuracy on just generated examples
        # entropy of the predicted label distribution, averaged equally across
        # both classes (if this drops then disc is very good or has given up)
        "disc_entropy": float(entropy),
        # true number of expert demos and predicted number of expert demos
        "disc_proportion_expert_true": float(pct_expert),
        "disc_proportion_expert_pred": float(pct_expert_pred),
        "n_expert": float(n_expert),
        "n_generated": float(n_generated),
    }


class AdversarialTrainer(base.DemonstrationAlgorithm[types.Transitions]):
    """Base class for adversarial imitation learning algorithms like GAIL and AIRL."""

    venv: vec_env.VecEnv
    """The original vectorized environment."""

    venv_train: vec_env.VecEnv
    """Like `self.venv`, but wrapped with train reward unless in debug mode.

    If `debug_use_ground_truth=True` was passed into the initializer then
    `self.venv_train` is the same as `self.venv`."""

    _demo_data_loader: Optional[Iterable[types.TransitionMapping]]
    _endless_expert_iterator: Optional[Iterator[types.TransitionMapping]]

    venv_wrapped: vec_env.VecEnvWrapper

    def __init__(
        self,
        *,
        demonstrations: base.AnyTransitions,
        demo_batch_size: int,
        venv: vec_env.VecEnv,
        gen_algo: base_class.BaseAlgorithm,
        reward_net: reward_nets.RewardNet,
        demo_minibatch_size: Optional[int] = None,
        n_disc_updates_per_round: int = 2,
        log_dir: types.AnyPath = "output/",
        disc_opt_cls: Type[th.optim.Optimizer] = th.optim.Adam,
        disc_opt_kwargs: Optional[Mapping] = None,
        gen_train_timesteps: Optional[int] = None,
        gen_replay_buffer_capacity: Optional[int] = None,
        custom_logger: Optional[logger.HierarchicalLogger] = None,
        init_tensorboard: bool = False,
        init_tensorboard_graph: bool = False,
        debug_use_ground_truth: bool = False,
        allow_variable_horizon: bool = False,
    ):
        """Builds AdversarialTrainer.

        Args:
            demonstrations: Demonstrations from an expert (optional). Transitions
                expressed directly as a `types.TransitionsMinimal` object, a sequence
                of trajectories, or an iterable of transition batches (mappings from
                keywords to arrays containing observations, etc).
            demo_batch_size: The number of samples in each batch of expert data. The
                discriminator batch size is twice this number because each discriminator
                batch contains a generator sample for every expert sample.
            venv: The vectorized environment to train in.
            gen_algo: The generator RL algorithm that is trained to maximize
                discriminator confusion. Environment and logger will be set to
                `venv` and `custom_logger`.
            reward_net: a Torch module that takes an observation, action and
                next observation tensors as input and computes a reward signal.
            demo_minibatch_size: size of minibatch to calculate gradients over.
                The gradients are accumulated until the entire batch is
                processed before making an optimization step. This is
                useful in GPU training to reduce memory usage, since
                fewer examples are loaded into memory at once,
                facilitating training with larger batch sizes, but is
                generally slower. Must be a factor of `demo_batch_size`.
                Optional, defaults to `demo_batch_size`.
            n_disc_updates_per_round: The number of discriminator updates after each
                round of generator updates in AdversarialTrainer.learn().
            log_dir: Directory to store TensorBoard logs, plots, etc. in.
            disc_opt_cls: The optimizer for discriminator training.
            disc_opt_kwargs: Parameters for discriminator training.
            gen_train_timesteps: The number of steps to train the generator policy for
                each iteration. If None, then defaults to the batch size (for on-policy)
                or number of environments (for off-policy).
            gen_replay_buffer_capacity: The capacity of the
                generator replay buffer (the number of obs-action-obs samples from
                the generator that can be stored). By default this is equal to
                `gen_train_timesteps`, meaning that we sample only from the most
                recent batch of generator samples.
            custom_logger: Where to log to; if None (default), creates a new logger.
            init_tensorboard: If True, makes various discriminator
                TensorBoard summaries.
            init_tensorboard_graph: If both this and `init_tensorboard` are True,
                then write a Tensorboard graph summary to disk.
            debug_use_ground_truth: If True, use the ground truth reward for
                `self.train_env`.
                This disables the reward wrapping that would normally replace
                the environment reward with the learned reward. This is useful for
                sanity checking that the policy training is functional.
            allow_variable_horizon: If False (default), algorithm will raise an
                exception if it detects trajectories of different length during
                training. If True, overrides this safety check. WARNING: variable
                horizon episodes leak information about the reward via termination
                condition, and can seriously confound evaluation. Read
                https://imitation.readthedocs.io/en/latest/guide/variable_horizon.html
                before overriding this.

        Raises:
            ValueError: if the batch size is not a multiple of the minibatch size.
        """
        self.demo_batch_size = demo_batch_size
        self.demo_minibatch_size = demo_minibatch_size or demo_batch_size
        if self.demo_batch_size % self.demo_minibatch_size != 0:
            raise ValueError("Batch size must be a multiple of minibatch size.")
        self._demo_data_loader = None
        self._endless_expert_iterator = None
        super().__init__(
            demonstrations=demonstrations,
            custom_logger=custom_logger,
            allow_variable_horizon=allow_variable_horizon,
        )

        self._global_step = 0
        self._disc_step = 0
        self.n_disc_updates_per_round = n_disc_updates_per_round

        self.debug_use_ground_truth = debug_use_ground_truth
        self.venv = venv
        self.gen_algo = gen_algo
        self._reward_net: reward_nets.RewardNet = reward_net.to(gen_algo.device)
        self._log_dir = util.parse_path(log_dir)

        # Create graph for optimising/recording stats on discriminator
        self._disc_opt_cls = disc_opt_cls
        self._disc_opt_kwargs = disc_opt_kwargs or {}
        self._init_tensorboard = init_tensorboard
        self._init_tensorboard_graph = init_tensorboard_graph
        self._disc_opt = self._disc_opt_cls(
            self._reward_net.parameters(),
            **self._disc_opt_kwargs,
        )

        if self._init_tensorboard:
            logging.info(f"building summary directory at {self._log_dir}")
            summary_dir = self._log_dir / "summary"
            summary_dir.mkdir(parents=True, exist_ok=True)
            self._summary_writer = thboard.SummaryWriter(str(summary_dir))

        self.venv_buffering = wrappers.BufferingWrapper(self.venv)

        if debug_use_ground_truth:
            # Would use an identity reward fn here, but RewardFns can't see rewards.
            self.venv_wrapped = self.venv_buffering
            self.gen_callback = None
        else:
            self.venv_wrapped = reward_wrapper.RewardVecEnvWrapper(
                self.venv_buffering,
                reward_fn=self.reward_train.predict_processed,
            )
            self.gen_callback = self.venv_wrapped.make_log_callback()
        self.venv_train = self.venv_wrapped

        self.gen_algo.set_env(self.venv_train)
        self.gen_algo.set_logger(self.logger)

        if gen_train_timesteps is None:
            gen_algo_env = self.gen_algo.get_env()
            assert gen_algo_env is not None
            self.gen_train_timesteps = gen_algo_env.num_envs
            if isinstance(self.gen_algo, on_policy_algorithm.OnPolicyAlgorithm):
                self.gen_train_timesteps *= self.gen_algo.n_steps
        else:
            self.gen_train_timesteps = gen_train_timesteps

        if gen_replay_buffer_capacity is None:
            gen_replay_buffer_capacity = self.gen_train_timesteps
        self._gen_replay_buffer = buffer.ReplayBuffer(
            gen_replay_buffer_capacity,
            self.venv,
        )

    @property
    def policy(self) -> policies.BasePolicy:
        policy = self.gen_algo.policy
        assert policy is not None
        return policy

    @abc.abstractmethod
    def logits_expert_is_high(
        self,
        state: th.Tensor,
        action: th.Tensor,
        next_state: th.Tensor,
        done: th.Tensor,
        log_policy_act_prob: Optional[th.Tensor] = None,
    ) -> th.Tensor:
        """Compute the discriminator's logits for each state-action sample.

        A high value corresponds to predicting expert, and a low value corresponds to
        predicting generator.

        Args:
            state: state at time t, of shape `(batch_size,) + state_shape`.
            action: action taken at time t, of shape `(batch_size,) + action_shape`.
            next_state: state at time t+1, of shape `(batch_size,) + state_shape`.
            done: binary episode completion flag after action at time t,
                of shape `(batch_size,)`.
            log_policy_act_prob: log probability of generator policy taking
                `action` at time t.

        Returns:
            Discriminator logits of shape `(batch_size,)`. A high output indicates an
            expert-like transition.
        """  # noqa: DAR202

    @property
    @abc.abstractmethod
    def reward_train(self) -> reward_nets.RewardNet:
        """Reward used to train generator policy."""

    @property
    @abc.abstractmethod
    def reward_test(self) -> reward_nets.RewardNet:
        """Reward used to train policy at "test" time after adversarial training."""

    def set_demonstrations(self, demonstrations: base.AnyTransitions) -> None:
        self._demo_data_loader = base.make_data_loader(
            demonstrations,
            self.demo_batch_size,
        )
        self._endless_expert_iterator = util.endless_iter(self._demo_data_loader)

    def _next_expert_batch(self) -> Mapping:
        assert self._endless_expert_iterator is not None
        return next(self._endless_expert_iterator)

    def train_disc(
        self,
        *,
        expert_samples: Optional[Mapping] = None,
        gen_samples: Optional[Mapping] = None,
    ) -> Mapping[str, float]:
        """Perform a single discriminator update, optionally using provided samples.

        Args:
            expert_samples: Transition samples from the expert in dictionary form.
                If provided, must contain keys corresponding to every field of the
                `Transitions` dataclass except "infos". All corresponding values can be
                either NumPy arrays or Tensors. Extra keys are ignored. Must contain
                `self.demo_batch_size` samples. If this argument is not provided, then
                `self.demo_batch_size` expert samples from `self.demo_data_loader` are
                used by default.
            gen_samples: Transition samples from the generator policy in same dictionary
                form as `expert_samples`. If provided, must contain exactly
                `self.demo_batch_size` samples. If not provided, then take
                `len(expert_samples)` samples from the generator replay buffer.

        Returns:
            Statistics for discriminator (e.g. loss, accuracy).
        """
        with self.logger.accumulate_means("disc"):
            # optionally write TB summaries for collected ops
            write_summaries = self._init_tensorboard and self._global_step % 20 == 0

            # compute loss
            self._disc_opt.zero_grad()

            batch_iter = self._make_disc_train_batches(
                gen_samples=gen_samples,
                expert_samples=expert_samples,
            )
            for batch in batch_iter:
                disc_logits = self.logits_expert_is_high(
                    batch["state"],
                    batch["action"],
                    batch["next_state"],
                    batch["done"],
                    batch["log_policy_act_prob"],
                )
                loss = F.binary_cross_entropy_with_logits(
                    disc_logits,
                    batch["labels_expert_is_one"].float(),
                )

                # Renormalise the loss to be averaged over the whole
                # batch size instead of the minibatch size.
                assert len(batch["state"]) == 2 * self.demo_minibatch_size
                loss *= self.demo_minibatch_size / self.demo_batch_size
                loss.backward()

            # do gradient step
            self._disc_opt.step()
            self._disc_step += 1

            # compute/write stats and TensorBoard data
            with th.no_grad():
                train_stats = compute_train_stats(
                    disc_logits,
                    batch["labels_expert_is_one"],
                    loss,
                )
            self.logger.record("global_step", self._global_step)
            for k, v in train_stats.items():
                self.logger.record(k, v)
            self.logger.dump(self._disc_step)
            if write_summaries:
                self._summary_writer.add_histogram("disc_logits", disc_logits.detach())

        return train_stats

    def train_gen(
        self,
        total_timesteps: Optional[int] = None,
        learn_kwargs: Optional[Mapping] = None,
    ) -> None:
        """Trains the generator to maximize the discriminator loss.

        After the end of training populates the generator replay buffer (used in
        discriminator training) with `self.disc_batch_size` transitions.

        Args:
            total_timesteps: The number of transitions to sample from
                `self.venv_train` during training. By default,
                `self.gen_train_timesteps`.
            learn_kwargs: kwargs for the Stable Baselines `RLModel.learn()`
                method.
        """
        if total_timesteps is None:
            total_timesteps = self.gen_train_timesteps
        if learn_kwargs is None:
            learn_kwargs = {}

        with self.logger.accumulate_means("gen"):
            self.gen_algo.learn(
                total_timesteps=total_timesteps,
                reset_num_timesteps=False,
                callback=self.gen_callback,
                **learn_kwargs,
            )
            self._global_step += 1

        gen_trajs, ep_lens = self.venv_buffering.pop_trajectories()
        self._check_fixed_horizon(ep_lens)
        gen_samples = rollout.flatten_trajectories_with_rew(gen_trajs)
        self._gen_replay_buffer.store(gen_samples)

    def train(
        self,
        total_timesteps: int,
        callback: Optional[Callable[[int], None]] = None,
    ) -> None:
        """Alternates between training the generator and discriminator.

        Every "round" consists of a call to `train_gen(self.gen_train_timesteps)`,
        a call to `train_disc`, and finally a call to `callback(round)`.

        Training ends once an additional "round" would cause the number of transitions
        sampled from the environment to exceed `total_timesteps`.

        Args:
            total_timesteps: An upper bound on the number of transitions to sample
                from the environment during training.
            callback: A function called at the end of every round which takes in a
                single argument, the round number. Round numbers are in
                `range(total_timesteps // self.gen_train_timesteps)`.
        """
        n_rounds = total_timesteps // self.gen_train_timesteps
        assert n_rounds >= 1, (
            "No updates (need at least "
            f"{self.gen_train_timesteps} timesteps, have only "
            f"total_timesteps={total_timesteps})!"
        )
        for r in tqdm.tqdm(range(0, n_rounds), desc="round"):
            self.train_gen(self.gen_train_timesteps)
            for _ in range(self.n_disc_updates_per_round):
                with networks.training(self.reward_train):
                    # switch to training mode (affects dropout, normalization)
                    self.train_disc()
            if callback:
                callback(r)
            self.logger.dump(self._global_step)

    @overload
    def _torchify_array(self, ndarray: np.ndarray) -> th.Tensor:
        ...

    @overload
    def _torchify_array(self, ndarray: None) -> None:
        ...

    def _torchify_array(self, ndarray: Optional[np.ndarray]) -> Optional[th.Tensor]:
        if ndarray is not None:
            return th.as_tensor(ndarray, device=self.reward_train.device)
        return None

    def _get_log_policy_act_prob(
        self,
        obs_th: th.Tensor,
        acts_th: th.Tensor,
    ) -> Optional[th.Tensor]:
        """Evaluates the given actions on the given observations.

        Args:
            obs_th: A batch of observations.
            acts_th: A batch of actions.

        Returns:
            A batch of log policy action probabilities.
        """
        if isinstance(self.policy, policies.ActorCriticPolicy):
            # policies.ActorCriticPolicy has a concrete implementation of
            # evaluate_actions to generate log_policy_act_prob given obs and actions.
            _, log_policy_act_prob_th, _ = self.policy.evaluate_actions(
                obs_th,
                acts_th,
            )
        elif isinstance(self.policy, sac_policies.SACPolicy):
            gen_algo_actor = self.policy.actor
            assert gen_algo_actor is not None
            # generate log_policy_act_prob from SAC actor.
            mean_actions, log_std, _ = gen_algo_actor.get_action_dist_params(obs_th)
            assert isinstance(
                gen_algo_actor.action_dist,
                distributions.SquashedDiagGaussianDistribution,
            )  # Note: this is just a hint to mypy
            distribution = gen_algo_actor.action_dist.proba_distribution(
                mean_actions,
                log_std,
            )
            # SAC applies a squashing function to bound the actions to a finite range
            # `acts_th` need to be scaled accordingly before computing log prob.
            # Scale actions only if the policy squashes outputs.
            assert self.policy.squash_output
            scaled_acts = self.policy.scale_action(acts_th.numpy(force=True))
            scaled_acts_th = th.as_tensor(scaled_acts, device=mean_actions.device)
            log_policy_act_prob_th = distribution.log_prob(scaled_acts_th)
        else:
            return None
        return log_policy_act_prob_th

    def _make_disc_train_batches(
        self,
        *,
        gen_samples: Optional[Mapping] = None,
        expert_samples: Optional[Mapping] = None,
    ) -> Iterator[Mapping[str, th.Tensor]]:
        """Build and return training minibatches for the next discriminator update.

        Args:
            gen_samples: Same as in `train_disc`.
            expert_samples: Same as in `train_disc`.

        Yields:
            The training minibatch: state, action, next state, dones, labels
            and policy log-probabilities.

        Raises:
            RuntimeError: Empty generator replay buffer.
            ValueError: `gen_samples` or `expert_samples` batch size is
                different from `self.demo_batch_size`.
        """
        batch_size = self.demo_batch_size

        if expert_samples is None:
            expert_samples = self._next_expert_batch()

        if gen_samples is None:
            if self._gen_replay_buffer.size() == 0:
                raise RuntimeError(
                    "No generator samples for training. " "Call `train_gen()` first.",
                )
            gen_samples_dataclass = self._gen_replay_buffer.sample(batch_size)
            gen_samples = types.dataclass_quick_asdict(gen_samples_dataclass)

        if not (len(gen_samples["obs"]) == len(expert_samples["obs"]) == batch_size):
            raise ValueError(
                "Need to have exactly `demo_batch_size` number of expert and "
                "generator samples, each. "
                f"(n_gen={len(gen_samples['obs'])} "
                f"n_expert={len(expert_samples['obs'])} "
                f"demo_batch_size={batch_size})",
            )

        # Guarantee that Mapping arguments are in mutable form.
        expert_samples = dict(expert_samples)
        gen_samples = dict(gen_samples)

        # Convert applicable Tensor values to NumPy.
        for field in dataclasses.fields(types.Transitions):
            k = field.name
            if k == "infos":
                continue
            for d in [gen_samples, expert_samples]:
                if isinstance(d[k], th.Tensor):
                    d[k] = d[k].detach().numpy()
        assert isinstance(gen_samples["obs"], np.ndarray)
        assert isinstance(expert_samples["obs"], np.ndarray)

        # Check dimensions.
        assert batch_size == len(expert_samples["acts"])
        assert batch_size == len(expert_samples["next_obs"])
        assert batch_size == len(gen_samples["acts"])
        assert batch_size == len(gen_samples["next_obs"])

        for start in range(0, batch_size, self.demo_minibatch_size):
            end = start + self.demo_minibatch_size
            # take minibatch slice (this creates views so no memory issues)
            expert_batch = {k: v[start:end] for k, v in expert_samples.items()}
            gen_batch = {k: v[start:end] for k, v in gen_samples.items()}

            # Concatenate rollouts, and label each row as expert or generator.
            obs = np.concatenate([expert_batch["obs"], gen_batch["obs"]])
            acts = np.concatenate([expert_batch["acts"], gen_batch["acts"]])
            next_obs = np.concatenate([expert_batch["next_obs"], gen_batch["next_obs"]])
            dones = np.concatenate([expert_batch["dones"], gen_batch["dones"]])
            # notice that the labels use the convention that expert samples are
            # labelled with 1 and generator samples with 0.
            labels_expert_is_one = np.concatenate(
                [
                    np.ones(self.demo_minibatch_size, dtype=int),
                    np.zeros(self.demo_minibatch_size, dtype=int),
                ],
            )

            # Calculate generator-policy log probabilities.
            with th.no_grad():
                obs_th = th.as_tensor(obs, device=self.gen_algo.device)
                acts_th = th.as_tensor(acts, device=self.gen_algo.device)
                log_policy_act_prob = self._get_log_policy_act_prob(obs_th, acts_th)
                if log_policy_act_prob is not None:
                    assert len(log_policy_act_prob) == 2 * self.demo_minibatch_size
                    log_policy_act_prob = log_policy_act_prob.reshape(
                        (2 * self.demo_minibatch_size,),
                    )
                del obs_th, acts_th  # unneeded

            obs_th, acts_th, next_obs_th, dones_th = self.reward_train.preprocess(
                obs,
                acts,
                next_obs,
                dones,
            )
            batch_dict = {
                "state": obs_th,
                "action": acts_th,
                "next_state": next_obs_th,
                "done": dones_th,
                "labels_expert_is_one": self._torchify_array(labels_expert_is_one),
                "log_policy_act_prob": log_policy_act_prob,
            }

            yield batch_dict



"""Evaluate policies: render policy interactively, save videos, log episode return."""

import logging
import pathlib
import time
from typing import Any, Mapping, Optional

import gymnasium as gym
import numpy as np
from sacred.observers import FileStorageObserver
from stable_baselines3.common.vec_env import VecEnvWrapper

from imitation.data import rollout, serialize
from imitation.policies.exploration_wrapper import ExplorationWrapper
from imitation.rewards import reward_wrapper
from imitation.rewards.serialize import load_reward
from imitation.scripts.config.eval_policy import eval_policy_ex
from imitation.scripts.ingredients import environment, expert
from imitation.scripts.ingredients import logging as logging_ingredient
from imitation.util import video_wrapper


class InteractiveRender(VecEnvWrapper):
    """Render the wrapped environment(s) on screen."""

    def __init__(self, venv, fps):
        """Builds renderer for `venv` running at `fps` frames per second."""
        super().__init__(venv)
        self.render_fps = fps

    def reset(self):
        ob = self.venv.reset()
        self.venv.render()
        return ob

    def step_wait(self):
        ob = self.venv.step_wait()
        if self.render_fps > 0:
            time.sleep(1 / self.render_fps)
        self.venv.render()
        return ob


def video_wrapper_factory(log_dir: pathlib.Path, **kwargs):
    """Returns a function that wraps the environment in a video recorder."""

    def f(env: gym.Env, i: int) -> video_wrapper.VideoWrapper:
        """Wraps `env` in a recorder saving videos to `{log_dir}/videos/{i}`."""
        directory = log_dir / "videos" / str(i)
        return video_wrapper.VideoWrapper(env, directory=directory, **kwargs)

    return f


@eval_policy_ex.main
def eval_policy(
    eval_n_timesteps: Optional[int],
    eval_n_episodes: Optional[int],
    render: bool,
    render_fps: int,
    videos: bool,
    video_kwargs: Mapping[str, Any],
    _run,
    _rnd: np.random.Generator,
    reward_type: Optional[str] = None,
    reward_path: Optional[str] = None,
    rollout_save_path: Optional[str] = None,
    explore_kwargs: Optional[Mapping[str, Any]] = None,
):
    """Rolls a policy out in an environment, collecting statistics.

    Args:
        eval_n_timesteps: Minimum number of timesteps to evaluate for. Set exactly
            one of `eval_n_episodes` and `eval_n_timesteps`.
        eval_n_episodes: Minimum number of episodes to evaluate for. Set exactly
            one of `eval_n_episodes` and `eval_n_timesteps`.
        render: If True, renders interactively to the screen.
        render_fps: The target number of frames per second to render on screen.
        videos: If True, saves videos to `log_dir`.
        video_kwargs: Keyword arguments passed through to `video_wrapper.VideoWrapper`.
        _rnd: Random number generator provided by Sacred.
        reward_type: If specified, overrides the environment reward with
            a reward of this.
        reward_path: If reward_type is specified, the path to a serialized reward
            of `reward_type` to override the environment reward with.
        rollout_save_path: where to save rollouts used for computing stats to disk;
            if None, then do not save.
        explore_kwargs: keyword arguments to an exploration wrapper to apply before
            rolling out, not including policy_callable, venv, and rng; if None, then
            do not wrap.

    Returns:
        Return value of `imitation.util.rollout.rollout_stats()`.
    """
    log_dir = logging_ingredient.make_log_dir()
    sample_until = rollout.make_sample_until(eval_n_timesteps, eval_n_episodes)
    post_wrappers = [video_wrapper_factory(log_dir, **video_kwargs)] if videos else None
    render_mode = "rgb_array" if videos else None
    with environment.make_venv(  # type: ignore[wrong-keyword-args]
        post_wrappers=post_wrappers,
        env_make_kwargs=dict(render_mode=render_mode),
    ) as venv:
        if render:
            venv = InteractiveRender(venv, render_fps)

        if reward_type is not None:
            reward_fn = load_reward(reward_type, reward_path, venv)
            venv = reward_wrapper.RewardVecEnvWrapper(venv, reward_fn)
            logging.info(f"Wrapped env in reward {reward_type} from {reward_path}.")

        policy = expert.get_expert_policy(venv)
        if explore_kwargs is not None:
            policy = ExplorationWrapper(
                policy,
                venv,
                rng=_rnd,
                **explore_kwargs,
            )
            log_str = (
                f"Wrapped policy in ExplorationWrapper with kwargs {explore_kwargs}"
            )
            logging.info(log_str)
        trajs = rollout.generate_trajectories(policy, venv, sample_until, rng=_rnd)

    if rollout_save_path:
        serialize.save(log_dir / rollout_save_path.replace("{log_dir}/", ""), trajs)

    return rollout.rollout_stats(trajs)


def main_console():
    observer_path = pathlib.Path.cwd() / "output" / "sacred" / "eval_policy"
    observer = FileStorageObserver(observer_path)
    eval_policy_ex.observers.append(observer)
    eval_policy_ex.run_commandline()


if __name__ == "__main__":  # pragma: no cover
    main_console()



"""Module of base classes and helper methods for imitation learning algorithms."""

import abc
from typing import (
    Any,
    Generic,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    TypeVar,
    Union,
    cast,
)

import torch.utils.data as th_data
from stable_baselines3.common import policies

from imitation.data import rollout, types
from imitation.util import logger as imit_logger
from imitation.util import util


class BaseImitationAlgorithm(abc.ABC):
    """Base class for all imitation learning algorithms."""

    _logger: imit_logger.HierarchicalLogger
    """Object to log statistics and natural language messages to."""

    allow_variable_horizon: bool
    """If True, allow variable horizon trajectories; otherwise error if detected."""

    _horizon: Optional[int]
    """Horizon of trajectories seen so far (None if no trajectories seen)."""

    def __init__(
        self,
        *,
        custom_logger: Optional[imit_logger.HierarchicalLogger] = None,
        allow_variable_horizon: bool = False,
    ):
        """Creates an imitation learning algorithm.

        Args:
            custom_logger: Where to log to; if None (default), creates a new logger.
            allow_variable_horizon: If False (default), algorithm will raise an
                exception if it detects trajectories of different length during
                training. If True, overrides this safety check. WARNING: variable
                horizon episodes leak information about the reward via termination
                condition, and can seriously confound evaluation. Read
                https://imitation.readthedocs.io/en/latest/getting-started/variable-horizon.html
                before overriding this.
        """
        self._logger = custom_logger or imit_logger.configure()
        self.allow_variable_horizon = allow_variable_horizon
        if allow_variable_horizon:
            self.logger.warn(
                "Running with `allow_variable_horizon` set to True. "
                "Some algorithms are biased towards shorter or longer "
                "episodes, which may significantly confound results. "
                "Additionally, even unbiased algorithms can exploit "
                "the information leak from the termination condition, "
                "producing spuriously high performance. See "
                "https://imitation.readthedocs.io/en/latest/getting-started/"
                "variable-horizon.html for more information.",
            )
        self._horizon = None

    @property
    def logger(self) -> imit_logger.HierarchicalLogger:
        return self._logger

    @logger.setter
    def logger(self, value: imit_logger.HierarchicalLogger) -> None:
        self._logger = value

    def _check_fixed_horizon(self, horizons: Iterable[int]) -> None:
        """Checks that episode lengths in `horizons` are fixed and equal to prior calls.

        If algorithm is safe to use with variable horizon episodes (e.g. behavioral
        cloning), then just don't call this method.

        Args:
            horizons: An iterable sequence of episode lengths.

        Raises:
            ValueError: The length of trajectories in trajs differs from one
                another, or from trajectory lengths in previous calls to this method.
        """
        if self.allow_variable_horizon:  # skip check -- YOLO
            return

        # horizons = all horizons seen so far (including trajs)
        horizons = set(horizons)
        if self._horizon is not None:
            horizons.add(self._horizon)

        if len(horizons) > 1:
            raise ValueError(
                f"Episodes of different length detected: {horizons}. "
                "Variable horizon environments are discouraged -- "
                "termination conditions leak information about reward. See "
                "https://imitation.readthedocs.io/en/latest/getting-started/"
                "variable-horizon.html for more information. "
                "If you are SURE you want to run imitation on a "
                "variable horizon task, then please pass in the flag: "
                "`allow_variable_horizon=True`.",
            )
        elif len(horizons) == 1:
            self._horizon = horizons.pop()

    def __getstate__(self):
        state = self.__dict__.copy()
        # logger can't be pickled as it depends on open files
        del state["_logger"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        # callee should modify self.logger directly if they want to override this
        self.logger = state.get("_logger") or imit_logger.configure()


TransitionKind = TypeVar("TransitionKind", bound=types.TransitionsMinimal)
AnyTransitions = Union[
    Iterable[types.Trajectory],
    Iterable[types.TransitionMapping],
    types.TransitionsMinimal,
]


class DemonstrationAlgorithm(BaseImitationAlgorithm, Generic[TransitionKind]):
    """An algorithm that learns from demonstration: BC, IRL, etc."""

    def __init__(
        self,
        *,
        demonstrations: Optional[AnyTransitions],
        custom_logger: Optional[imit_logger.HierarchicalLogger] = None,
        allow_variable_horizon: bool = False,
    ):
        """Creates an algorithm that learns from demonstrations.

        Args:
            demonstrations: Demonstrations from an expert (optional). Transitions
                expressed directly as a `types.TransitionsMinimal` object, a sequence
                of trajectories, or an iterable of transition batches (mappings from
                keywords to arrays containing observations, etc).
            custom_logger: Where to log to; if None (default), creates a new logger.
            allow_variable_horizon: If False (default), algorithm will raise an
                exception if it detects trajectories of different length during
                training. If True, overrides this safety check. WARNING: variable
                horizon episodes leak information about the reward via termination
                condition, and can seriously confound evaluation. Read
                https://imitation.readthedocs.io/en/latest/getting-started/variable-horizon.html
                before overriding this.
        """
        super().__init__(
            custom_logger=custom_logger,
            allow_variable_horizon=allow_variable_horizon,
        )

        if demonstrations is not None:
            self.set_demonstrations(demonstrations)

    @abc.abstractmethod
    def set_demonstrations(self, demonstrations: AnyTransitions) -> None:
        """Sets the demonstration data.

        Changing the demonstration data on-demand can be useful for
        interactive algorithms like DAgger.

        Args:
             demonstrations: Either a Torch `DataLoader`, any other iterator that
                yields dictionaries containing "obs" and "acts" Tensors or NumPy arrays,
                `TransitionKind` instance, or a Sequence of Trajectory objects.
        """

    @property
    @abc.abstractmethod
    def policy(self) -> policies.BasePolicy:
        """Returns a policy imitating the demonstration data."""


class _WrappedDataLoader:
    """Wraps a data loader (batch iterable) and checks for specified batch size."""

    def __init__(
        self,
        data_loader: Iterable[types.TransitionMapping],
        expected_batch_size: int,
    ):
        """Builds _WrappedDataLoader.

        Args:
            data_loader: The data loader (batch iterable) to wrap.
            expected_batch_size: The batch size to check for.
        """
        self.data_loader = data_loader
        self.expected_batch_size = expected_batch_size

    def __iter__(self) -> Iterator[types.TransitionMapping]:
        """Yields data from `self.data_loader`, checking `self.expected_batch_size`.

        Yields:
            Identity -- yields same batches as from `self.data_loader`.

        Raises:
            ValueError: `self.data_loader` returns a batch of size not equal to
                `self.expected_batch_size`.
        """
        for batch in self.data_loader:
            if len(batch["obs"]) != self.expected_batch_size:
                raise ValueError(
                    f"Expected batch size {self.expected_batch_size} "
                    f"!= {len(batch['obs'])} = len(batch['obs'])",
                )
            if len(batch["acts"]) != self.expected_batch_size:
                raise ValueError(
                    f"Expected batch size {self.expected_batch_size} "
                    f"!= {len(batch['acts'])} = len(batch['acts'])",
                )
            yield batch


def make_data_loader(
    transitions: AnyTransitions,
    batch_size: int,
    data_loader_kwargs: Optional[Mapping[str, Any]] = None,
) -> Iterable[types.TransitionMapping]:
    """Converts demonstration data to Torch data loader.

    Args:
        transitions: Transitions expressed directly as a `types.TransitionsMinimal`
            object, a sequence of trajectories, or an iterable of transition
            batches (mappings from keywords to arrays containing observations, etc).
        batch_size: The size of the batch to create. Does not change the batch size
            if `transitions` is already an iterable of transition batches.
        data_loader_kwargs: Arguments to pass to `th_data.DataLoader`.

    Returns:
        An iterable of transition batches.

    Raises:
        ValueError: if `transitions` is an iterable over transition batches with batch
            size not equal to `batch_size`; or if `transitions` is transitions or a
            sequence of trajectories with total timesteps less than `batch_size`.
        TypeError: if `transitions` is an unsupported type.
    """
    if batch_size <= 0:
        raise ValueError(f"batch_size={batch_size} must be positive.")

    if isinstance(transitions, Iterable):
        # Inferring the correct type here is difficult with generics.
        (
            first_item,
            transitions,
        ) = util.get_first_iter_element(  # type: ignore[assignment]
            transitions,
        )
        if isinstance(first_item, types.Trajectory):
            transitions = cast(Iterable[types.Trajectory], transitions)
            transitions = rollout.flatten_trajectories(list(transitions))

    if isinstance(transitions, types.TransitionsMinimal):
        if len(transitions) < batch_size:
            raise ValueError(
                f"Number of transitions in `demonstrations` {len(transitions)} "
                f"is smaller than batch size {batch_size}.",
            )

        kwargs: Mapping[str, Any] = {
            "shuffle": True,
            "drop_last": True,
            **(data_loader_kwargs or {}),
        }
        return th_data.DataLoader(
            transitions,
            batch_size=batch_size,
            collate_fn=types.transitions_collate_fn,
            **kwargs,
        )
    elif isinstance(transitions, Iterable):
        # Safe to ignore this error since we've already converted Iterable[Trajectory]
        # `transitions` into Iterable[TransitionMapping]
        return _WrappedDataLoader(transitions, batch_size)  # type: ignore[arg-type]
    else:
        raise TypeError(f"`demonstrations` unexpected type {type(transitions)}")



"""Tests `imitation.util.reward_wrapper`."""

import numpy as np

from imitation.data import rollout
from imitation.policies.base import RandomPolicy
from imitation.rewards import reward_wrapper
from imitation.util import util


class FunkyReward:
    """A reward that ignores observation and depends only on batch index."""

    def __call__(self, obs, act, next_obs, steps=None):
        """Returns consecutive reward from 1 to batch size `len(obs)`."""
        # give each environment number from 1 to num_envs
        return (np.arange(len(obs)) + 1).astype("float32")


def test_reward_overwrite(rng):
    """Test that reward wrapper actually overwrites base rewards."""
    env_name = "Pendulum-v1"
    num_envs = 3
    env = util.make_vec_env(env_name, rng=rng, n_envs=num_envs)
    reward_fn = FunkyReward()
    wrapped_env = reward_wrapper.RewardVecEnvWrapper(env, reward_fn)
    policy = RandomPolicy(env.observation_space, env.action_space)
    sample_until = rollout.make_min_episodes(10)
    default_stats = rollout.rollout_stats(
        rollout.generate_trajectories(policy, env, sample_until, rng),
    )
    wrapped_stats = rollout.rollout_stats(
        rollout.generate_trajectories(policy, wrapped_env, sample_until, rng),
    )
    # Pendulum-v1 always has negative rewards
    assert default_stats["return_max"] < 0
    # ours gives between 1 * traj_len and num_envs * traj_len reward
    # (trajectories are all constant length of 200 in Pendulum)
    steps = wrapped_stats["len_mean"]
    assert wrapped_stats["return_min"] == 1 * steps
    assert wrapped_stats["return_max"] == num_envs * steps

    # check that wrapped reward is negative (all pendulum rewards is negative)
    # and other rewards are non-negative
    rand_act, _ = policy.predict(wrapped_env.reset())
    _, rew, _, infos = wrapped_env.step(rand_act)
    assert np.all(rew >= 0)
    assert np.all([info_dict["original_env_rew"] < 0 for info_dict in infos])



"""Generative Adversarial Imitation Learning (GAIL)."""

from typing import Optional

import torch as th
from stable_baselines3.common import base_class, vec_env
from torch.nn import functional as F

from imitation.algorithms import base
from imitation.algorithms.adversarial import common
from imitation.rewards import reward_nets


class RewardNetFromDiscriminatorLogit(reward_nets.RewardNet):
    r"""Converts the discriminator logits raw value to a reward signal.

    Wrapper for reward network that takes in the logits of the discriminator
    probability distribution and outputs the corresponding reward for the GAIL
    algorithm.

    Below is the derivation of the transformation that needs to be applied.

    The GAIL paper defines the cost function of the generator as:

    .. math::

        \log{D}

    as shown on line 5 of Algorithm 1. In the paper, :math:`D` is the probability
    distribution learned by the discriminator, where :math:`D(X)=1` if the trajectory
    comes from the generator, and :math:`D(X)=0` if it comes from the expert.
    In this implementation, we have decided to use the opposite convention that
    :math:`D(X)=0` if the trajectory comes from the generator,
    and :math:`D(X)=1` if it comes from the expert. Therefore, the resulting cost
    function is:

    .. math::

        \log{(1-D)}

    Since our algorithm trains using a reward function instead of a loss function, we
    need to invert the sign to get:

    .. math::

        R=-\log{(1-D)}=\log{\frac{1}{1-D}}

    Now, let :math:`L` be the output of our reward net, which gives us the logits of D
    (:math:`L=\operatorname{logit}{D}`). We can write:

    .. math::

        D=\operatorname{sigmoid}{L}=\frac{1}{1+e^{-L}}

    Since :math:`1-\operatorname{sigmoid}{(L)}` is the same as
    :math:`\operatorname{sigmoid}{(-L)}`, we can write:

    .. math::

        R=-\log{\operatorname{sigmoid}{(-L)}}

    which is a non-decreasing map from the logits of D to the reward.
    """

    def __init__(self, base: reward_nets.RewardNet):
        """Builds LogSigmoidRewardNet to wrap `reward_net`."""
        # TODO(adam): make an explicit RewardNetWrapper class?
        super().__init__(
            observation_space=base.observation_space,
            action_space=base.action_space,
            normalize_images=base.normalize_images,
        )
        self.base = base

    def forward(
        self,
        state: th.Tensor,
        action: th.Tensor,
        next_state: th.Tensor,
        done: th.Tensor,
    ) -> th.Tensor:
        logits = self.base.forward(state, action, next_state, done)
        return -F.logsigmoid(-logits)


class GAIL(common.AdversarialTrainer):
    """Generative Adversarial Imitation Learning (`GAIL`_).

    .. _GAIL: https://arxiv.org/abs/1606.03476
    """

    def __init__(
        self,
        *,
        demonstrations: base.AnyTransitions,
        demo_batch_size: int,
        venv: vec_env.VecEnv,
        gen_algo: base_class.BaseAlgorithm,
        reward_net: reward_nets.RewardNet,
        **kwargs,
    ):
        """Generative Adversarial Imitation Learning.

        Args:
            demonstrations: Demonstrations from an expert (optional). Transitions
                expressed directly as a `types.TransitionsMinimal` object, a sequence
                of trajectories, or an iterable of transition batches (mappings from
                keywords to arrays containing observations, etc).
            demo_batch_size: The number of samples in each batch of expert data. The
                discriminator batch size is twice this number because each discriminator
                batch contains a generator sample for every expert sample.
            venv: The vectorized environment to train in.
            gen_algo: The generator RL algorithm that is trained to maximize
                discriminator confusion. Environment and logger will be set to
                `venv` and `custom_logger`.
            reward_net: a Torch module that takes an observation, action and
                next observation tensor as input, then computes the logits.
                Used as the GAIL discriminator.
            **kwargs: Passed through to `AdversarialTrainer.__init__`.
        """
        # Raw self._reward_net is discriminator logits
        reward_net = reward_net.to(gen_algo.device)
        # Process it to produce output suitable for RL training
        # Applies a -log(sigmoid(-logits)) to the logits (see class for explanation)
        self._processed_reward = RewardNetFromDiscriminatorLogit(reward_net)
        super().__init__(
            demonstrations=demonstrations,
            demo_batch_size=demo_batch_size,
            venv=venv,
            gen_algo=gen_algo,
            reward_net=reward_net,
            **kwargs,
        )

    def logits_expert_is_high(
        self,
        state: th.Tensor,
        action: th.Tensor,
        next_state: th.Tensor,
        done: th.Tensor,
        log_policy_act_prob: Optional[th.Tensor] = None,
    ) -> th.Tensor:
        r"""Compute the discriminator's logits for each state-action sample.

        Args:
            state: The state of the environment at the time of the action.
            action: The action taken by the expert or generator.
            next_state: The state of the environment after the action.
            done: whether a `terminal state` (as defined under the MDP of the task) has
                been reached.
            log_policy_act_prob: The log probability of the action taken by the
                generator, :math:`\log{P(a|s)}`.

        Returns:
            The logits of the discriminator for each state-action sample.
        """
        del log_policy_act_prob
        logits = self._reward_net(state, action, next_state, done)
        assert logits.shape == state.shape[:1]
        return logits

    @property
    def reward_train(self) -> reward_nets.RewardNet:
        return self._processed_reward

    @property
    def reward_test(self) -> reward_nets.RewardNet:
        return self._processed_reward



"""Generate a markdown summary of the results of a benchmarking run."""
import argparse
import pathlib
from collections import Counter
from functools import lru_cache
from typing import Generator, Sequence, cast

import datasets
import numpy as np
from huggingface_sb3 import EnvironmentName
from rliable import library as rly
from rliable import metrics

from imitation.data import rollout, types
from imitation.data.huggingface_utils import TrajectoryDatasetSequence
from imitation.util.sacred_file_parsing import (
    find_sacred_runs,
    group_runs_by_algo_and_env,
)


@lru_cache(maxsize=None)
def get_random_agent_score(env: str):
    stats = rollout.rollout_stats(
        cast(
            Sequence[types.TrajectoryWithRew],
            TrajectoryDatasetSequence(
                datasets.load_dataset(
                    f"HumanCompatibleAI/random-{EnvironmentName(env)}",
                )["train"],
            ),
        ),
    )
    return stats["monitor_return_mean"]


def print_markdown_summary(path: pathlib.Path) -> Generator[str, None, None]:
    if not path.exists():
        raise NotADirectoryError(f"Path {path} does not exist.")

    yield "# Benchmark Summary"
    yield ""
    yield (
        f"This is a summary of the sacred runs in `{path}` generated by "
        f"`sacred_output_to_markdown_summary.py`."
    )

    runs_by_algo_and_env = group_runs_by_algo_and_env(path)
    algos = sorted(runs_by_algo_and_env.keys())

    status_counts = Counter((run["status"] for _, run in find_sacred_runs(path)))
    statuses = sorted(list(status_counts))
    # Note: we only print the status section if there are multiple statuses
    if not (len(statuses) == 1 and statuses[0] == "COMPLETED"):
        yield "## Run status" ""
        yield "Status | Count"
        yield "--- | ---"
        for status in statuses:
            yield f"{status} | {status_counts[status]}"
        yield ""

        yield "## Detailed Run Status"
        yield f"Algorithm | Environment | {' | '.join(statuses)}"
        yield "--- | --- " + " | --- " * len(statuses)
        for algo in algos:
            envs = sorted(runs_by_algo_and_env[algo].keys())
            for env in envs:
                status_counts = Counter(
                    (run["status"] for run in runs_by_algo_and_env[algo][env]),
                )
                yield (
                    f"{algo} | {env} | "
                    f"{' | '.join([str(status_counts[status]) for status in statuses])}"
                )

    yield "## Scores"
    yield ""
    yield (
        "The scores are normalized based on the performance of a random agent as the"
        " baseline and the expert as the maximum possible score as explained "
        "[in this blog post](https://araffin.github.io/post/rliable/):"
    )
    yield "> `(score - random_score) / (expert_score - random_score)`"
    yield ""
    yield (
        "Aggregate scores and confidence intervals are computed using the "
        "[rliable library](https://agarwl.github.io/rliable/)."
    )

    for algo in algos:
        yield f"### {algo.upper()}"
        yield "Environment | Score (mean/std)| Normalized Score (mean/std) | N"
        yield " --- | --- | --- | --- "
        envs = sorted(runs_by_algo_and_env[algo].keys())
        accumulated_normalized_scores = []
        for env in envs:
            scores = [
                run["result"]["imit_stats"]["monitor_return_mean"]
                for run in runs_by_algo_and_env[algo][env]
            ]
            expert_scores = [
                run["result"]["expert_stats"]["monitor_return_mean"]
                for run in runs_by_algo_and_env[algo][env]
            ]
            random_score = get_random_agent_score(env)
            normalized_score = [
                (score - random_score) / (expert_score - random_score)
                for score, expert_score in zip(scores, expert_scores)
            ]
            accumulated_normalized_scores.append(normalized_score)

            yield (
                f"{env} | "
                f"{np.mean(scores):.3f} / {np.std(scores):.3f} | "
                f"{np.mean(normalized_score):.3f} / {np.std(normalized_score):.3f} | "
                f"{len(scores)}"
            )

        aggregate_scores, aggregate_score_cis = rly.get_interval_estimates(
            {"normalized_score": np.asarray(accumulated_normalized_scores).T},
            lambda x: np.array([metrics.aggregate_mean(x), metrics.aggregate_iqm(x)]),
            reps=1000,
        )
        yield ""
        yield "#### Aggregate Normalized scores"

        yield "Metric | Value | 95% CI"
        yield " --- | --- | --- "
        yield (
            f"Mean | "
            f"{aggregate_scores['normalized_score'][0]:.3f} | "
            f"[{aggregate_score_cis['normalized_score'][0][0]:.3f}, "
            f"{aggregate_score_cis['normalized_score'][0][1]:.3f}]"
        )
        yield (
            f"IQM | "
            f"{aggregate_scores['normalized_score'][1]:.3f} | "
            f"[{aggregate_score_cis['normalized_score'][1][0]:.3f}, "
            f"{aggregate_score_cis['normalized_score'][1][1]:.3f}]"
        )
        yield ""


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Generate a markdown summary of the results of a benchmarking run.",
    )
    parser.add_argument("path", type=pathlib.Path)
    parser.add_argument("--output", type=pathlib.Path, default="summary.md")

    args = parser.parse_args()

    with open(args.output, "w") as fh:
        for line in print_markdown_summary(pathlib.Path(args.path)):
            fh.write(line)
            fh.write("\n")
            fh.flush()



"""This ingredient provides a reinforcement learning algorithm from stable-baselines3.

The algorithm instance is either freshly constructed or loaded from a file.
"""

import logging
import warnings
from typing import Any, Dict, Mapping, Optional, Type

import sacred
import stable_baselines3 as sb3
from stable_baselines3.common import (
    base_class,
    buffers,
    off_policy_algorithm,
    on_policy_algorithm,
    vec_env,
)

from imitation.policies import serialize
from imitation.policies.replay_buffer_wrapper import ReplayBufferRewardWrapper
from imitation.rewards.reward_function import RewardFn
from imitation.scripts.ingredients import logging as logging_ingredient
from imitation.scripts.ingredients.policy import policy_ingredient

rl_ingredient = sacred.Ingredient(
    "rl",
    ingredients=[policy_ingredient, logging_ingredient.logging_ingredient],
)
logger = logging.getLogger(__name__)


@rl_ingredient.config
def config():
    rl_cls = None
    batch_size = None
    rl_kwargs = dict()
    locals()  # quieten flake8


@rl_ingredient.config_hook
def config_hook(config, command_name, logger):
    """Sets defaults equivalent to sb3.PPO default hyperparameters.

    This hook is a no-op if command_name is "sqil" (used only in train_imitation),
    which has its own config hook.

    Args:
        config: Sacred config dict.
        command_name: Sacred command name.
        logger: Sacred logger.

    Returns:
        config: Updated Sacred config dict.
    """
    del logger
    res = {}
    if config["rl"]["rl_cls"] is None and command_name != "sqil":
        res["rl_cls"] = sb3.PPO
        res["batch_size"] = 2048  # rl_kwargs["n_steps"] = batch_size // venv.num_envs
        res["rl_kwargs"] = dict(
            learning_rate=3e-4,
            batch_size=64,
            n_epochs=10,
            ent_coef=0.0,
        )
    return res


@rl_ingredient.named_config
def fast():
    batch_size = 2
    rl_kwargs = dict(
        # SB3 RL seems to need batch size of 2, otherwise it runs into numeric
        # issues when computing multinomial distribution during predict()
        batch_size=2,
        # Setting n_epochs=1 speeds up thing a lot
        n_epochs=1,
    )
    locals()  # quieten flake8


@rl_ingredient.named_config
def sac():
    # For recommended SAC hyperparams in each environment, see:
    # https://github.com/DLR-RM/rl-baselines3-zoo/blob/master/hyperparams/sac.yml
    rl_cls = sb3.SAC
    warnings.warn(
        "SAC currently only supports continuous action spaces. "
        "Consider adding a discrete version as mentioned here: "
        "https://github.com/DLR-RM/stable-baselines3/issues/505",
        category=RuntimeWarning,
    )
    # Default HPs are as follows:
    batch_size = 256  # batch size for RL algorithm
    rl_kwargs = dict(batch_size=None)  # make sure to set batch size to None

    locals()  # quieten flake8


def _maybe_add_relabel_buffer(
    rl_kwargs: Dict[str, Any],
    relabel_reward_fn: Optional[RewardFn] = None,
) -> Dict[str, Any]:
    """Use ReplayBufferRewardWrapper in rl_kwargs if relabel_reward_fn is not None."""
    rl_kwargs = dict(rl_kwargs)
    if relabel_reward_fn:
        _buffer_kwargs = dict(reward_fn=relabel_reward_fn)
        _buffer_kwargs["replay_buffer_class"] = rl_kwargs.get(
            "replay_buffer_class",
            buffers.ReplayBuffer,
        )
        rl_kwargs["replay_buffer_class"] = ReplayBufferRewardWrapper

        if "replay_buffer_kwargs" in rl_kwargs:
            _buffer_kwargs.update(rl_kwargs["replay_buffer_kwargs"])
        rl_kwargs["replay_buffer_kwargs"] = _buffer_kwargs
    return rl_kwargs


@rl_ingredient.capture
def make_rl_algo(
    venv: vec_env.VecEnv,
    rl_cls: Type[base_class.BaseAlgorithm],
    batch_size: int,
    rl_kwargs: Mapping[str, Any],
    policy: Mapping[str, Any],
    _seed: int,
    relabel_reward_fn: Optional[RewardFn] = None,
) -> base_class.BaseAlgorithm:
    """Instantiates a Stable Baselines3 RL algorithm.

    Args:
        venv: The vectorized environment to train on.
        rl_cls: Type of a Stable Baselines3 RL algorithm.
        batch_size: The batch size of the RL algorithm.
        rl_kwargs: Keyword arguments for RL algorithm constructor.
        policy: Configuration for the policy ingredient. We need the
            policy_cls and policy_kwargs component.
        relabel_reward_fn: Reward function used for reward relabeling
            in replay or rollout buffers of RL algorithms.

    Returns:
        The RL algorithm.

    Raises:
        ValueError: `gen_batch_size` not divisible by `venv.num_envs`.
        TypeError: `rl_cls` is neither `OnPolicyAlgorithm` nor `OffPolicyAlgorithm`.
    """
    if batch_size % venv.num_envs != 0:
        raise ValueError(
            f"num_envs={venv.num_envs} must evenly divide batch_size={batch_size}.",
        )
    rl_kwargs = dict(rl_kwargs)

    # TODO: this is a hack and an indicator that the rl ingredient should be refactored
    if rl_cls == sb3.SAC:
        del rl_kwargs["n_epochs"]

    # If on-policy, collect `batch_size` many timesteps each update.
    # If off-policy, train on `batch_size` many timesteps each update.
    # These are different notion of batches, but this seems the closest
    # possible translation, and I would expect the appropriate hyperparameter
    # to be similar between them.
    if issubclass(rl_cls, on_policy_algorithm.OnPolicyAlgorithm):
        assert "n_steps" not in rl_kwargs, (
            "set 'n_steps' at top-level using 'batch_size'. "
            "n_steps = batch_size // num_vec"
        )
        rl_kwargs["n_steps"] = batch_size // venv.num_envs
    elif issubclass(rl_cls, off_policy_algorithm.OffPolicyAlgorithm):
        if rl_kwargs.get("batch_size") is not None:
            raise ValueError("set 'batch_size' at top-level")
        rl_kwargs["batch_size"] = batch_size
        rl_kwargs = _maybe_add_relabel_buffer(
            rl_kwargs=rl_kwargs,
            relabel_reward_fn=relabel_reward_fn,
        )
    else:
        raise TypeError(f"Unsupported RL algorithm '{rl_cls}'")
    rl_algo = rl_cls(
        policy=policy["policy_cls"],
        # Note(yawen): Copy `policy_kwargs` as SB3 may mutate the config we pass.
        # In particular, policy_kwargs["use_sde"] may be changed in rl_cls.__init__()
        # for certain algorithms, such as Soft Actor Critic. See:
        # https://github.com/DLR-RM/stable-baselines3/blob/30772aa9f53a4cf61571ee90046cdc454c1b11d7/sb3/common/off_policy_algorithm.py#L145
        policy_kwargs=dict(policy["policy_kwargs"]),
        env=venv,
        seed=_seed,
        **rl_kwargs,
    )
    logger.info(f"RL algorithm: {type(rl_algo)}")
    logger.info(f"Policy network summary:\n {rl_algo.policy}")
    return rl_algo


@rl_ingredient.capture
def load_rl_algo_from_path(
    _seed: int,
    agent_path: str,
    venv: vec_env.VecEnv,
    rl_cls: Type[base_class.BaseAlgorithm],
    rl_kwargs: Mapping[str, Any],
    relabel_reward_fn: Optional[RewardFn] = None,
) -> base_class.BaseAlgorithm:
    rl_kwargs = dict(rl_kwargs)

    # TODO: this is a hack and an indicator that the rl ingredient should be refactored
    if rl_cls == sb3.SAC:
        del rl_kwargs["n_epochs"]

    if issubclass(rl_cls, off_policy_algorithm.OffPolicyAlgorithm):
        rl_kwargs = _maybe_add_relabel_buffer(
            rl_kwargs=rl_kwargs,
            relabel_reward_fn=relabel_reward_fn,
        )
    agent = serialize.load_stable_baselines_model(
        cls=rl_cls,
        path=agent_path,
        venv=venv,
        seed=_seed,
        **rl_kwargs,
    )
    logger.info(f"Warm starting agent from '{agent_path}'")
    logger.info(f"RL algorithm: {type(agent)}")
    logger.info(f"Policy network summary:\n {agent.policy}")

    return agent



"""Tests `imitation.rewards.reward_nets` and `imitation.rewards.serialize`."""

import functools
import logging
import os
import tempfile
from typing import Callable, Tuple
from unittest import mock

import gymnasium as gym
import numpy as np
import pytest
import torch as th
from stable_baselines3.common.vec_env import DummyVecEnv

import imitation.testing.reward_nets as testing_reward_nets
from imitation.data import rollout
from imitation.policies import base
from imitation.rewards import reward_nets, serialize
from imitation.util import networks, util


def _potential(x):
    # _potential is never actually called in the tests: we just need a dummy
    # potential to be able to construct shaped reward networks.
    return th.zeros(x.shape[0], device=x.device)  # pragma: no cover


ENVS = ["FrozenLake-v1", "CartPole-v1", "Pendulum-v1"]
IMAGE_ENVS = ["Asteroids-v4"]
DESERIALIZATION_TYPES = [
    "zero",
    "RewardNet_normalized",
    "RewardNet_unnormalized",
    "RewardNet_shaped",
    "RewardNet_unshaped",
]


# Reward net classes, allowed kwargs
MAKE_REWARD_NET = [
    functools.partial(
        reward_nets.BasicRewardNet,
        hid_sizes=(3,),
    ),
    functools.partial(
        reward_nets.BasicShapedRewardNet,
        reward_hid_sizes=(3,),
        potential_hid_sizes=(3,),
    ),
    functools.partial(
        testing_reward_nets.make_ensemble,
        hid_sizes=(3,),
    ),
]
MAKE_IMAGE_REWARD_NET = [
    functools.partial(
        reward_nets.CnnRewardNet,
        hid_channels=(3, 3, 3),
        kernel_size=3,
        padding=0,
        stride=3,
    ),
]

MakePredictProcessedWrapper = Callable[
    [reward_nets.RewardNet],
    reward_nets.PredictProcessedWrapper,
]
MAKE_PREDICT_PROCESSED_WRAPPERS = [
    lambda base: reward_nets.NormalizedRewardNet(base, networks.RunningNorm),
]

MakeForwardWrapper = Callable[[reward_nets.RewardNet], reward_nets.ForwardWrapper]
MAKE_FORWARD_WRAPPERS = [
    lambda base: reward_nets.ShapedRewardNet(base, _potential, 0.99),
]

REWARD_NET_KWARGS = [
    {},
    {"normalize_input_layer": networks.RunningNorm},
    {"normalize_input_layer": networks.EMANorm},
    {"use_next_state": True, "dropout_prob": 0.3},
    {"use_done": True},
]

IMAGE_REWARD_NET_KWARGS = [
    {},
    {"use_next_state": True, "use_action": False},
    {"dropout_prob": 0.3},
    {"use_done": True, "use_action": True},
    {"use_done": True, "use_action": False},
]

NORMALIZE_OUTPUT_LAYER = [
    None,
    networks.RunningNorm,
]

NumpyTransitions = Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]


@pytest.fixture
def numpy_transitions() -> NumpyTransitions:
    """A batch of states, actions, next_states, and dones as np.ndarrays for Env2D."""
    return (
        np.zeros((10, 5, 5)),
        np.zeros((10, 1), dtype=int),
        np.zeros((10, 5, 5)),
        np.zeros((10,), dtype=bool),
    )


TorchTransitions = Tuple[th.Tensor, th.Tensor, th.Tensor, th.Tensor]


@pytest.fixture
def torch_transitions() -> TorchTransitions:
    """A batch of states, actions, next_states, and dones as th.Tensors for Env2D."""
    return (
        th.zeros((10, 5, 5)),
        th.zeros((10, 1), dtype=th.int),
        th.zeros((10, 5, 5)),
        th.zeros((10,), dtype=th.bool),
    )


def _init_no_crash(env_name, reward_net_cls, reward_net_kwargs, normalize_output_layer):
    env = gym.make(env_name)

    reward_net = reward_net_cls(
        env.observation_space,
        env.action_space,
        **reward_net_kwargs,
    )

    if normalize_output_layer:
        reward_net = reward_nets.NormalizedRewardNet(
            reward_net,
            normalize_output_layer,
        )


@pytest.mark.parametrize("env_name", ENVS)
@pytest.mark.parametrize("reward_net_cls", MAKE_REWARD_NET)
@pytest.mark.parametrize("reward_net_kwargs", REWARD_NET_KWARGS)
@pytest.mark.parametrize("normalize_output_layer", NORMALIZE_OUTPUT_LAYER)
def test_init_no_crash(
    env_name,
    reward_net_cls,
    reward_net_kwargs,
    normalize_output_layer,
):
    _init_no_crash(
        env_name,
        reward_net_cls,
        reward_net_kwargs,
        normalize_output_layer,
    )


@pytest.mark.parametrize("env_name", IMAGE_ENVS)
@pytest.mark.parametrize("reward_net_cls", MAKE_IMAGE_REWARD_NET)
@pytest.mark.parametrize("reward_net_kwargs", IMAGE_REWARD_NET_KWARGS)
@pytest.mark.parametrize("normalize_output_layer", NORMALIZE_OUTPUT_LAYER)
def test_image_init_no_crash(
    env_name,
    reward_net_cls,
    reward_net_kwargs,
    normalize_output_layer,
):
    _init_no_crash(
        env_name,
        reward_net_cls,
        reward_net_kwargs,
        normalize_output_layer,
    )


NOT_QUITE_IMAGE_SPACES = (
    gym.spaces.Box(0, 255, shape=(210, 160, 3), dtype=np.float32),
    gym.spaces.Box(0, 254, shape=(210, 160, 3), dtype=np.uint8),
    gym.spaces.Box(1, 255, shape=(210, 160, 3), dtype=np.uint8),
    gym.spaces.Box(0, 255, shape=(210, 160), dtype=np.uint8),
)


def test_cnn_reward_net_input_validation():
    atari_obs_space = gym.spaces.Box(0, 255, shape=(210, 160, 3), dtype=np.uint8)
    atari_action_space = gym.spaces.Discrete(14)

    with pytest.raises(ValueError, match="must take current or next state"):
        reward_nets.CnnRewardNet(
            atari_obs_space,
            atari_action_space,
            use_state=False,
            use_next_state=False,
        )

    for obs_space in NOT_QUITE_IMAGE_SPACES:
        with pytest.raises(ValueError, match="requires observations to be images"):
            reward_nets.CnnRewardNet(
                obs_space,
                atari_action_space,
            )

    illegal_act_space = gym.spaces.MultiDiscrete((2, 2))
    with pytest.raises(ValueError, match="can only use Discrete action spaces."):
        reward_nets.CnnRewardNet(
            atari_obs_space,
            illegal_act_space,
        )

    reward_nets.CnnRewardNet(
        atari_obs_space,
        illegal_act_space,
        use_action=False,
    )


def test_cnn_potential_input_validation():
    for obs_space in NOT_QUITE_IMAGE_SPACES:
        with pytest.raises(ValueError, match="must be given image inputs"):
            reward_nets.BasicPotentialCNN(
                obs_space,
                hid_sizes=(32,),
            )


@pytest.mark.parametrize("dimensions", (1, 3, 4, 5))
def test_cnn_transpose_input_validation(dimensions: int):
    shape = (2,) * dimensions
    tens = th.zeros(shape)

    if dimensions == 4:  # should succeed
        reward_nets.cnn_transpose(tens)
    else:  # should fail
        with pytest.raises(ValueError, match="Invalid input: "):
            reward_nets.cnn_transpose(tens)


def _sample(space, n):
    return np.array([space.sample() for _ in range(n)])


def _make_env_and_save_reward_net(env_name, reward_type, tmpdir, rng, is_image=False):
    venv = util.make_vec_env(
        env_name,
        n_envs=1,
        parallel=False,
        rng=rng,
    )
    save_path = os.path.join(tmpdir, "norm_reward.pt")

    assert reward_type in [
        "zero",
        "RewardNet_normalized",
        "RewardNet_unnormalized",
        "RewardNet_shaped",
        "RewardNet_unshaped",
    ], f"Reward net type {reward_type} not supported by this helper."

    if reward_type == "zero":
        return venv, save_path

    net_cls = reward_nets.CnnRewardNet if is_image else reward_nets.BasicRewardNet
    net = net_cls(venv.observation_space, venv.action_space)

    if reward_type == "RewardNet_normalized":
        net = reward_nets.NormalizedRewardNet(net, networks.RunningNorm)
    elif reward_type == "RewardNet_shaped":
        pot_cls = (
            reward_nets.BasicPotentialMLP
            if not is_image
            else reward_nets.BasicPotentialCNN
        )
        potential = pot_cls(venv.observation_space, [8, 8])
        net = reward_nets.ShapedRewardNet(net, potential, discount_factor=0.99)
    elif reward_type in ["RewardNet_unshaped", "RewardNet_unnormalized"]:
        pass

    th.save(net, save_path)
    return venv, save_path


def _is_reward_valid(env_name, reward_type, tmpdir, rng, is_image):
    venv, tmppath = _make_env_and_save_reward_net(
        env_name,
        reward_type,
        tmpdir,
        rng,
        is_image=is_image,
    )

    TRAJECTORY_LEN = 10
    obs = _sample(venv.observation_space, TRAJECTORY_LEN)
    actions = _sample(venv.action_space, TRAJECTORY_LEN)
    next_obs = _sample(venv.observation_space, TRAJECTORY_LEN)
    steps = np.arange(0, TRAJECTORY_LEN)

    reward_fn = serialize.load_reward(reward_type, tmppath, venv)
    pred_reward = reward_fn(obs, actions, next_obs, steps)

    assert isinstance(pred_reward, np.ndarray)
    assert pred_reward.shape == (TRAJECTORY_LEN,)
    assert np.issubdtype(pred_reward.dtype, np.number)


@pytest.mark.parametrize("env_name", ENVS)
@pytest.mark.parametrize("reward_type", DESERIALIZATION_TYPES)
def test_reward_valid(env_name, reward_type, tmpdir, rng):
    """Test output of reward function is appropriate shape and type."""
    _is_reward_valid(env_name, reward_type, tmpdir, rng, is_image=False)


@pytest.mark.parametrize("env_name", IMAGE_ENVS)
@pytest.mark.parametrize("reward_type", DESERIALIZATION_TYPES)
def test_reward_valid_image(env_name, reward_type, tmpdir, rng):
    """Test output of reward function is appropriate shape and type."""
    _is_reward_valid(env_name, reward_type, tmpdir, rng, is_image=True)


def test_strip_wrappers_basic(rng):
    venv = util.make_vec_env(
        "FrozenLake-v1",
        n_envs=1,
        parallel=False,
        rng=rng,
    )
    net = reward_nets.BasicRewardNet(venv.observation_space, venv.action_space)
    net = reward_nets.NormalizedRewardNet(net, networks.RunningNorm)
    net = serialize._strip_wrappers(
        net,
        wrapper_types=[reward_nets.NormalizedRewardNet],
    )
    assert isinstance(net, reward_nets.BasicRewardNet)
    # This removing a wrapper from an unwrapped reward net should do nothing
    net = serialize._strip_wrappers(net, wrapper_types=[reward_nets.ShapedRewardNet])
    assert isinstance(net, reward_nets.BasicRewardNet)


def test_strip_wrappers_image_basic(rng):
    venv = util.make_vec_env("Asteroids-v4", n_envs=1, parallel=False, rng=rng)
    net = reward_nets.CnnRewardNet(venv.observation_space, venv.action_space)
    net = reward_nets.NormalizedRewardNet(net, networks.RunningNorm)
    net = serialize._strip_wrappers(
        net,
        wrapper_types=[reward_nets.NormalizedRewardNet],
    )
    assert isinstance(net, reward_nets.CnnRewardNet)
    # This removing a wrapper from an unwrapped reward net should do nothing
    net = serialize._strip_wrappers(net, wrapper_types=[reward_nets.ShapedRewardNet])
    assert isinstance(net, reward_nets.CnnRewardNet)


def test_strip_wrappers_complex(rng):
    venv = util.make_vec_env(
        "FrozenLake-v1",
        n_envs=1,
        parallel=False,
        rng=rng,
    )
    net = reward_nets.BasicRewardNet(venv.observation_space, venv.action_space)
    net = reward_nets.ShapedRewardNet(net, _potential, discount_factor=0.99)
    net = reward_nets.NormalizedRewardNet(net, networks.RunningNorm)
    # Removing in incorrect order should do nothing
    net = serialize._strip_wrappers(
        net,
        wrapper_types=[reward_nets.ShapedRewardNet, reward_nets.NormalizedRewardNet],
    )

    assert isinstance(net, reward_nets.NormalizedRewardNet)
    assert isinstance(net.base, reward_nets.ShapedRewardNet)
    # Correct order should work
    net = serialize._strip_wrappers(
        net,
        wrapper_types=[reward_nets.NormalizedRewardNet, reward_nets.ShapedRewardNet],
    )
    assert isinstance(net, reward_nets.BasicRewardNet)


def test_strip_wrappers_image_complex(rng):
    venv = util.make_vec_env("Asteroids-v4", n_envs=1, parallel=False, rng=rng)
    net = reward_nets.CnnRewardNet(venv.observation_space, venv.action_space)
    net = reward_nets.ShapedRewardNet(net, _potential, discount_factor=0.99)
    net = reward_nets.NormalizedRewardNet(net, networks.RunningNorm)
    # Removing in incorrect order should do nothing
    net = serialize._strip_wrappers(
        net,
        wrapper_types=[reward_nets.ShapedRewardNet, reward_nets.NormalizedRewardNet],
    )

    assert isinstance(net, reward_nets.NormalizedRewardNet)
    assert isinstance(net.base, reward_nets.ShapedRewardNet)
    # Correct order should work
    net = serialize._strip_wrappers(
        net,
        wrapper_types=[reward_nets.NormalizedRewardNet, reward_nets.ShapedRewardNet],
    )
    assert isinstance(net, reward_nets.CnnRewardNet)


def test_validate_wrapper_structure():
    env = gym.make("FrozenLake-v1")

    class RewardNetA(reward_nets.RewardNet):
        def forward(*args):
            ...  # pragma: no cover

    class WrapperB(reward_nets.RewardNetWrapper):
        def forward(*args):
            ...  # pragma: no cover

    reward_net = RewardNetA(env.action_space, env.observation_space)
    reward_net = WrapperB(reward_net)

    assert isinstance(reward_net.base, RewardNetA)

    # This should not raise a type error
    serialize._validate_wrapper_structure(reward_net, {(WrapperB, RewardNetA)})

    raises_error_ctxmgr = pytest.raises(
        TypeError,
        match=r"Wrapper structure should match \[.*\] but found \[.*\]",
    )

    # The top level wrapper is an instance of WrapperB this should raise a type error
    with raises_error_ctxmgr:
        serialize._validate_wrapper_structure(reward_net, {(RewardNetA,)})

    # Reward net is not wrapped at all this should raise a type error.
    with raises_error_ctxmgr:
        serialize._validate_wrapper_structure(
            RewardNetA(env.action_space, env.observation_space),
            {(WrapperB,)},
        )

    # The prefix is longer then set of wrappers
    with raises_error_ctxmgr:
        serialize._validate_wrapper_structure(
            reward_net,
            {(WrapperB, RewardNetA, WrapperB)},
        )

    # This should not raise a type error since one of the prefixes matches
    serialize._validate_wrapper_structure(
        reward_net,
        {(WrapperB, RewardNetA), (RewardNetA,)},
    )

    # This should raise a type error since none the prefix is in the incorrect order
    with raises_error_ctxmgr:
        serialize._validate_wrapper_structure(reward_net, {(RewardNetA, WrapperB)})


@pytest.mark.parametrize("env_name", ENVS)
def test_cant_load_unnorm_as_norm(env_name, tmpdir, rng):
    venv, tmppath = _make_env_and_save_reward_net(
        env_name,
        "RewardNet_unnormalized",
        tmpdir,
        rng=rng,
    )
    with pytest.raises(TypeError):
        serialize.load_reward("RewardNet_normalized", tmppath, venv)


def _serialize_deserialize_identity(
    env_name,
    net_cls,
    net_kwargs,
    normalize_rewards,
    tmpdir,
    rng,
):
    """Does output of deserialized reward network match that of original?"""
    logging.info(f"Testing {net_cls}")

    ep_steps = 5

    venv = util.make_vec_env(
        env_name,
        n_envs=1,
        parallel=False,
        rng=rng,
        max_episode_steps=ep_steps + 1,
    )
    original = net_cls(venv.observation_space, venv.action_space, **net_kwargs)
    if normalize_rewards:
        original = reward_nets.NormalizedRewardNet(original, networks.RunningNorm)
    random = base.RandomPolicy(venv.observation_space, venv.action_space)

    tmppath = os.path.join(tmpdir, "reward.pt")
    th.save(original, tmppath)
    loaded = th.load(tmppath)

    assert original.observation_space == loaded.observation_space
    assert original.action_space == loaded.action_space

    transitions = rollout.generate_transitions(
        random,
        venv,
        n_timesteps=ep_steps,
        rng=rng,
    )

    if isinstance(original, reward_nets.NormalizedRewardNet):
        wrapped_rew_fn = serialize.load_reward("RewardNet_normalized", tmppath, venv)
        unwrapped_rew_fn = serialize.load_reward(
            "RewardNet_unnormalized",
            tmppath,
            venv,
        )
    elif isinstance(original, reward_nets.ShapedRewardNet):
        unwrapped_rew_fn = serialize.load_reward("RewardNet_unshaped", tmppath, venv)
        wrapped_rew_fn = serialize.load_reward("RewardNet_shaped", tmppath, venv)
    else:
        unwrapped_rew_fn = serialize.load_reward("RewardNet_unshaped", tmppath, venv)
        wrapped_rew_fn = unwrapped_rew_fn

    rewards = {
        "wrapped": [],
        "unwrapped": [],
    }
    for net in [original, loaded]:
        trans_args = (
            transitions.obs,
            transitions.acts,
            transitions.next_obs,
            transitions.dones,
        )
        rewards["wrapped"].append(net.predict(*trans_args))
        if hasattr(net, "base"):
            rewards["unwrapped"].append(net.base.predict(*trans_args))
        else:
            rewards["unwrapped"].append(net.predict(*trans_args))

    args = (
        transitions.obs,
        transitions.acts,
        transitions.next_obs,
        transitions.dones,
    )
    rewards["wrapped"].append(wrapped_rew_fn(*args))
    rewards["unwrapped"].append(unwrapped_rew_fn(*args))

    for key, predictions in rewards.items():
        assert len(predictions) == 3
        assert np.allclose(predictions[0], predictions[1])
        assert np.allclose(predictions[0], predictions[2])


@pytest.mark.parametrize("env_name", ENVS)
@pytest.mark.parametrize("net_cls", MAKE_REWARD_NET)
@pytest.mark.parametrize("net_kwargs", REWARD_NET_KWARGS)
@pytest.mark.parametrize("normalize_rewards", [True, False])
def test_serialize_identity(
    env_name,
    net_cls,
    net_kwargs,
    normalize_rewards,
    tmpdir,
    rng,
):
    """Does output of deserialized reward MLP match that of original?"""
    _serialize_deserialize_identity(
        env_name,
        net_cls,
        net_kwargs,
        normalize_rewards,
        tmpdir,
        rng,
    )


@pytest.mark.parametrize("env_name", IMAGE_ENVS)
@pytest.mark.parametrize("net_cls", MAKE_IMAGE_REWARD_NET)
@pytest.mark.parametrize("net_kwargs", IMAGE_REWARD_NET_KWARGS)
@pytest.mark.parametrize("normalize_rewards", [True, False])
def test_serialize_identity_images(
    env_name,
    net_cls,
    net_kwargs,
    normalize_rewards,
    tmpdir,
    rng,
):
    """Does output of deserialized reward CNN match that of original?"""
    _serialize_deserialize_identity(
        env_name,
        net_cls,
        net_kwargs,
        normalize_rewards,
        tmpdir,
        rng,
    )


class Env2D(gym.Env):
    """Mock environment with 2D observations."""

    def __init__(self):
        """Builds `Env2D`."""
        super().__init__()
        self.observation_space = gym.spaces.Box(shape=(5, 5), low=-1.0, high=1.0)
        self.action_space = gym.spaces.Discrete(2)

    def step(self, action):
        obs = self.observation_space.sample()
        rew = 0.0
        done = False
        info = {}
        return obs, rew, done, info

    def reset(self):
        return self.observation_space.sample()


def test_potential_net_2d_obs():
    """Test potential net can do forward-prop with 2D observation.

    This is a regression test for a problem identified Eric. Previously, reward
    nets would not properly flatten N-dimensional states before passing them to
    potential networks, leading to shape mismatches.
    """
    # instantiate environment & get batch observations, actions, etc.
    env = Env2D()
    obs = env.reset()
    action = env.action_space.sample()
    next_obs, _, done, _ = env.step(action)
    obs_b = obs[None]
    action_b = np.array([action], dtype="int")
    next_obs_b = next_obs[None]
    done_b = np.array([done], dtype="bool")

    net = reward_nets.BasicShapedRewardNet(
        env.observation_space,
        env.action_space,
    )
    rew_batch = net.predict(obs_b, action_b, next_obs_b, done_b)
    assert rew_batch.shape == (1,)


@pytest.fixture
def env_2d() -> Env2D:
    """An instance of Env2d."""
    return Env2D()


def test_ensemble_errors_if_there_are_too_few_members(env_2d):
    for num_members in range(2):
        with pytest.raises(ValueError):
            reward_nets.RewardEnsemble(
                env_2d.observation_space,
                env_2d.action_space,
                members=[
                    testing_reward_nets.MockRewardNet(
                        env_2d.observation_space,
                        env_2d.action_space,
                    )
                    for _ in range(num_members)
                ],
            )


@pytest.fixture
def zero_reward_net(env_2d) -> testing_reward_nets.MockRewardNet:
    return testing_reward_nets.MockRewardNet(
        env_2d.observation_space,
        env_2d.action_space,
        value=0,
    )


@pytest.fixture(params=MAKE_PREDICT_PROCESSED_WRAPPERS)
def predict_processed_wrapper(request, zero_reward_net):
    return request.param(zero_reward_net)


@pytest.fixture
def two_ensemble(env_2d) -> reward_nets.RewardEnsemble:
    """A simple reward ensemble made up of two mock reward nets."""
    return reward_nets.RewardEnsemble(
        env_2d.observation_space,
        env_2d.action_space,
        members=[
            testing_reward_nets.MockRewardNet(
                env_2d.observation_space,
                env_2d.action_space,
            )
            for _ in range(2)
        ],
    )


def test_reward_ensemble_predict_reward_moments(
    two_ensemble: reward_nets.RewardEnsemble,
    numpy_transitions: NumpyTransitions,
):
    # Test that the calculation of mean and variance is correct
    two_ensemble.members[0].value = 0
    two_ensemble.members[1].value = 0
    mean, var = two_ensemble.predict_reward_moments(*numpy_transitions)
    assert np.isclose(mean, 0).all()
    assert np.isclose(var, 0).all()
    two_ensemble.members[0].value = 3
    two_ensemble.members[1].value = -1
    mean, var = two_ensemble.predict_reward_moments(*numpy_transitions)
    assert np.isclose(mean, 1).all()
    assert np.isclose(var, 8).all()  # note we are using the unbiased variance estimator
    # Test that ensemble calls members correctly
    two_ensemble.members[0].forward = mock.MagicMock(return_value=th.zeros(10))
    mean, var = two_ensemble.predict_reward_moments(*numpy_transitions)
    two_ensemble.members[0].forward.assert_called_once()


def test_ensemble_members_have_different_parameters(env_2d):
    ensemble = testing_reward_nets.make_ensemble(
        env_2d.observation_space,
        env_2d.action_space,
    )

    assert not th.allclose(
        next(ensemble.members[0].parameters()),
        next(ensemble.members[1].parameters()),
    )


def test_add_std_wrapper_raises_error_when_wrapping_wrong_type(env_2d):
    mock_env = testing_reward_nets.MockRewardNet(
        env_2d.observation_space,
        env_2d.action_space,
    )
    assert not isinstance(mock_env, reward_nets.RewardNetWithVariance)
    with pytest.raises(TypeError):
        reward_nets.AddSTDRewardWrapper(mock_env, default_alpha=0.1)


def test_add_std_reward_wrapper(
    two_ensemble: reward_nets.RewardEnsemble,
    numpy_transitions: NumpyTransitions,
):
    two_ensemble.members[0].value = 3
    two_ensemble.members[1].value = -1
    reward_fn = reward_nets.AddSTDRewardWrapper(two_ensemble, default_alpha=0.1)
    rewards = reward_fn.predict_processed(*numpy_transitions)
    assert np.allclose(rewards, 1 + 0.1 * np.sqrt(8))
    # test overriding in predict processed works correctly
    rewards = reward_fn.predict_processed(*numpy_transitions, alpha=-0.5)
    assert np.allclose(rewards, 1 - 0.5 * np.sqrt(8))


def test_shaped_reward_net(
    zero_reward_net: testing_reward_nets.MockRewardNet,
    numpy_transitions: NumpyTransitions,
):
    def potential(x: th.Tensor):
        return th.full((x.shape[0],), 10, device=x.device)

    shaped = reward_nets.ShapedRewardNet(zero_reward_net, potential, 0.9)
    # We expect the shaped reward to be -1 since,
    # r'(s,a,s') = r(s,a,s') + \gamma \theta(s') - \theta(s) = (0) + (0.9)(10) - 10 = -1
    shaped_rew = th.full((10,), -1, dtype=th.float32)
    forward_args = shaped.preprocess(*numpy_transitions)
    assert th.allclose(shaped(*forward_args), shaped_rew)
    assert th.allclose(shaped.predict_th(*numpy_transitions), shaped_rew)
    assert np.allclose(shaped.predict(*numpy_transitions), shaped_rew.numpy())
    assert np.allclose(shaped.predict_processed(*numpy_transitions), shaped_rew.numpy())


@pytest.mark.parametrize("make_forward_wrapper", MAKE_FORWARD_WRAPPERS)
def test_forward_wrapper_cannot_be_applied_predict_processed_wrapper(
    predict_processed_wrapper: reward_nets.PredictProcessedWrapper,
    make_forward_wrapper: MakeForwardWrapper,
):
    with pytest.raises(
        ValueError,
        match=r"ForwardWrapper cannot be applied on top of PredictProcessedWrapper!",
    ):
        make_forward_wrapper(predict_processed_wrapper)


@pytest.mark.parametrize(
    "make_predict_processed_wrapper",
    MAKE_PREDICT_PROCESSED_WRAPPERS,
)
def test_predict_processed_wrappers_pass_on_kwargs(
    make_predict_processed_wrapper: MakePredictProcessedWrapper,
    zero_reward_net: testing_reward_nets.MockRewardNet,
    numpy_transitions: NumpyTransitions,
):
    zero_reward_net.predict_processed = mock.Mock(  # type: ignore[assignment]
        return_value=np.zeros((10,)),
    )
    wrapped_reward_net = make_predict_processed_wrapper(
        zero_reward_net,
    )
    wrapped_reward_net.predict_processed(
        *numpy_transitions,
        foobar=42,
    )
    zero_reward_net.predict_processed.assert_called_once_with(
        *numpy_transitions,
        foobar=42,
    )


def test_predict_processed_wrappers_to_pass_on_method_calls_to_base(
    numpy_transitions: NumpyTransitions,
    torch_transitions: TorchTransitions,
    predict_processed_wrapper: reward_nets.PredictProcessedWrapper,
):
    base = mock.create_autospec(testing_reward_nets.MockRewardNet)

    base.device = th.device("cpu")

    base.dtype = th.float32

    predict_processed_wrapper._base = base
    # Check method calls
    for attr, call_with, return_value in [
        ("forward", torch_transitions, th.zeros(10)),
        ("predict_th", numpy_transitions, np.zeros(10)),
        ("predict", numpy_transitions, np.zeros(10)),
        ("predict_processed", numpy_transitions, np.zeros(10)),
        ("preprocess", numpy_transitions, torch_transitions),
    ]:
        setattr(base, attr, mock.MagicMock(return_value=return_value))
        getattr(predict_processed_wrapper, attr)(*call_with)
        getattr(base, attr).assert_called_once_with(*call_with)

    # Check property lookups
    assert predict_processed_wrapper.device is base.device

    assert predict_processed_wrapper.dtype is base.dtype


def test_load_reward_passes_along_alpha_to_add_std_wrappers_predict_processed_method(
    env_2d: Env2D,
    two_ensemble: reward_nets.RewardEnsemble,
    numpy_transitions: NumpyTransitions,
):
    """Kwargs passed to load_reward are passed along to predict_processed."""
    two_ensemble.members[0].value = 3
    two_ensemble.members[1].value = -1
    reward_net = reward_nets.AddSTDRewardWrapper(two_ensemble, default_alpha=0)
    with tempfile.TemporaryDirectory() as tmp_dir:
        net_path = os.path.join(tmp_dir, "reward_net.pkl")
        th.save(reward_net, os.path.join(net_path))
        new_alpha = -0.5
        reward_fn = serialize.load_reward(
            "RewardNet_std_added",
            net_path,
            env_2d,
            alpha=new_alpha,
        )
        rewards = reward_fn(*numpy_transitions)
        assert np.allclose(rewards, 1 + new_alpha * np.sqrt(8))


@pytest.mark.parametrize("env_name", ENVS)
def test_device_for_parameterless_model(env_name):
    class ParameterlessNet(reward_nets.RewardNet):
        def forward(self):
            """Dummy function to avoid abstractmethod complaints."""

    env = gym.make(env_name)
    net = ParameterlessNet(env.observation_space, env.action_space)
    assert net.device == th.device("cpu")


@pytest.mark.parametrize("normalize_input_layer", [None, networks.RunningNorm])
def test_training_regression(normalize_input_layer, rng):
    """Test reward_net normalization by training a regression model."""
    venv = DummyVecEnv([lambda: gym.make("CartPole-v0")] * 2)
    reward_net = reward_nets.BasicRewardNet(
        venv.observation_space,
        venv.action_space,
        normalize_input_layer=normalize_input_layer,
    )
    norm_rew_net = reward_nets.NormalizedRewardNet(
        reward_net,
        normalize_output_layer=networks.RunningNorm,
    )

    # Construct a loss function and an Optimizer.
    criterion = th.nn.MSELoss(reduction="sum")
    optimizer = th.optim.SGD(norm_rew_net.parameters(), lr=1e-6)

    # Getting transitions from a random policy
    random = base.RandomPolicy(venv.observation_space, venv.action_space)
    for _ in range(2):
        transitions = rollout.generate_transitions(
            random,
            venv,
            n_timesteps=100,
            rng=rng,
        )
        trans_args = (
            transitions.obs,
            transitions.acts,
            transitions.next_obs,
            transitions.dones,
        )
        trans_args_th = norm_rew_net.preprocess(*trans_args)
        rews_th = norm_rew_net(*trans_args_th)
        rews = rews_th.detach().cpu().numpy().flatten()

        # Compute and print loss
        loss = criterion(
            util.safe_to_tensor(transitions.rews).to(norm_rew_net.device),
            rews_th,
        )

        # Get rewards from norm_rew_net.predict() and norm_rew_net.predict_processed()
        rews_predict = norm_rew_net.predict(*trans_args)
        rews_processed = norm_rew_net.predict_processed(*trans_args)

        # norm_rew_net() and norm_rew_net.predict() don't pass the reward through the
        # normalization layer, so the values of `rews` and `rews_predict` are identical
        assert (rews == rews_predict).all()
        # norm_rew_net.predict_processed() does normalize the reward
        assert not (rews_processed == rews_predict).all()

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()



"""Density-based baselines for imitation learning.

Each of these algorithms learns a density estimate on some aspect of the demonstrations,
then rewards the agent for following that estimate.
"""

import enum
import itertools
from collections.abc import Mapping
from typing import Any, Dict, Iterable, List, Optional, cast

import numpy as np
from gymnasium.spaces import utils as space_utils
from sklearn import neighbors, preprocessing
from stable_baselines3.common import base_class, vec_env

from imitation.algorithms import base
from imitation.data import rollout, types, wrappers
from imitation.rewards import reward_wrapper
from imitation.util import logger as imit_logger
from imitation.util import util


class DensityType(enum.Enum):
    """Input type the density model should use."""

    STATE_DENSITY = enum.auto()
    """Density on state s."""

    STATE_ACTION_DENSITY = enum.auto()
    """Density on (s,a) pairs."""

    STATE_STATE_DENSITY = enum.auto()
    """Density on (s,s') pairs."""


class DensityAlgorithm(base.DemonstrationAlgorithm):
    """Learns a reward function based on density modeling.

    Specifically, it constructs a non-parametric estimate of `p(s)`, `p(s,a)`, `p(s,s')`
    and then computes a reward using the log of these probabilities.
    """

    is_stationary: bool
    density_type: DensityType
    venv: vec_env.VecEnv
    transitions: Dict[Optional[int], np.ndarray]
    kernel: str
    kernel_bandwidth: float
    standardise: bool

    _scaler: Optional[preprocessing.StandardScaler]
    _density_models: Dict[Optional[int], neighbors.KernelDensity]
    rl_algo: Optional[base_class.BaseAlgorithm]
    buffering_wrapper: wrappers.BufferingWrapper
    venv_wrapped: reward_wrapper.RewardVecEnvWrapper
    wrapper_callback: reward_wrapper.WrappedRewardCallback

    def __init__(
        self,
        *,
        demonstrations: Optional[base.AnyTransitions],
        venv: vec_env.VecEnv,
        rng: np.random.Generator,
        density_type: DensityType = DensityType.STATE_ACTION_DENSITY,
        kernel: str = "gaussian",
        kernel_bandwidth: float = 0.5,
        rl_algo: Optional[base_class.BaseAlgorithm] = None,
        is_stationary: bool = True,
        standardise_inputs: bool = True,
        custom_logger: Optional[imit_logger.HierarchicalLogger] = None,
        allow_variable_horizon: bool = False,
    ):
        """Builds DensityAlgorithm.

        Args:
            demonstrations: expert demonstration trajectories.
            density_type: type of density to train on: single state, state-action pairs,
                or state-state pairs.
            kernel: kernel to use for density estimation with `sklearn.KernelDensity`.
            kernel_bandwidth: bandwidth of kernel. If `standardise_inputs` is
                true and you are using a Gaussian kernel, then it probably makes sense
                to set this somewhere between 0.1 and 1.
            venv: The environment to learn a reward model in. We don't actually need
                any environment interaction to fit the reward model, but we use this
                to extract the observation and action space, and to train the RL
                algorithm `rl_algo` (if specified).
            rng: random state for sampling from demonstrations.
            rl_algo: An RL algorithm to train on the resulting reward model (optional).
            is_stationary: if True, share same density models for all timesteps;
                if False, use a different density model for each timestep.
                A non-stationary model is particularly likely to be useful when using
                STATE_DENSITY, to encourage agent to imitate entire trajectories, not
                just a few states that have high frequency in the demonstration dataset.
                If non-stationary, demonstrations must be trajectories, not transitions
                (which do not contain timesteps).
            standardise_inputs: if True, then the inputs to the reward model
                will be standardised to have zero mean and unit variance over the
                demonstration trajectories. Otherwise, inputs will be passed to the
                reward model with their ordinary scale.
            custom_logger: Where to log to; if None (default), creates a new logger.
            allow_variable_horizon: If False (default), algorithm will raise an
                exception if it detects trajectories of different length during
                training. If True, overrides this safety check. WARNING: variable
                horizon episodes leak information about the reward via termination
                condition, and can seriously confound evaluation. Read
                https://imitation.readthedocs.io/en/latest/guide/variable_horizon.html
                before overriding this.
        """
        self.is_stationary = is_stationary
        self.density_type = density_type
        self.venv = venv
        self.transitions = dict()
        super().__init__(
            demonstrations=demonstrations,
            custom_logger=custom_logger,
            allow_variable_horizon=allow_variable_horizon,
        )

        self.kernel = kernel
        self.kernel_bandwidth = kernel_bandwidth
        self.standardise = standardise_inputs
        self._scaler = None
        self._density_models = dict()
        self.rng = rng

        self.rl_algo = rl_algo
        self.buffering_wrapper = wrappers.BufferingWrapper(self.venv)
        self.venv_wrapped = reward_wrapper.RewardVecEnvWrapper(
            self.buffering_wrapper,
            self,
        )
        self.wrapper_callback = self.venv_wrapped.make_log_callback()

    def _get_demo_from_batch(
        self,
        obs_b: types.Observation,
        act_b: np.ndarray,
        next_obs_b: Optional[types.Observation],
    ) -> Dict[Optional[int], List[np.ndarray]]:
        if next_obs_b is None and self.density_type == DensityType.STATE_STATE_DENSITY:
            raise ValueError(
                "STATE_STATE_DENSITY requires next_obs_b "
                "to be provided, but it was None",
            )

        assert act_b.shape[1:] == self.venv.action_space.shape
        ob_space = self.venv.observation_space
        if isinstance(obs_b, types.DictObs):
            exp_shape = {
                k: v.shape for k, v in ob_space.items()  # type: ignore[attr-defined]
            }
            obs_shape = {k: v.shape[1:] for k, v in obs_b.items()}
            assert exp_shape == obs_shape, f"Expected {exp_shape}, got {obs_shape}"
        else:
            assert obs_b.shape[1:] == ob_space.shape
        assert len(act_b) == len(obs_b)
        if next_obs_b is not None:
            assert next_obs_b.shape == obs_b.shape

        if next_obs_b is not None:
            next_obs_b_iterator: Iterable = next_obs_b
        else:
            next_obs_b_iterator = itertools.repeat(None)

        transitions: Dict[Optional[int], List[np.ndarray]] = {}
        for obs, act, next_obs in zip(obs_b, act_b, next_obs_b_iterator):
            flat_trans = self._preprocess_transition(obs, act, next_obs)
            transitions.setdefault(None, []).append(flat_trans)
        return transitions

    def set_demonstrations(self, demonstrations: base.AnyTransitions) -> None:
        """Sets the demonstration data."""
        transitions: Dict[Optional[int], List[np.ndarray]] = {}

        if isinstance(demonstrations, types.TransitionsMinimal):
            next_obs_b = getattr(demonstrations, "next_obs", None)
            transitions.update(
                self._get_demo_from_batch(
                    demonstrations.obs,
                    demonstrations.acts,
                    next_obs_b,
                ),
            )
        elif isinstance(demonstrations, Iterable):
            # Inferring the correct type here is difficult with generics.
            (
                first_item,
                demonstrations,
            ) = util.get_first_iter_element(  # type: ignore[assignment]
                demonstrations,
            )
            if isinstance(first_item, types.Trajectory):
                # we assume that all elements are also types.Trajectory.
                # (this means we have timestamp information)
                # It's not perfectly type safe, but it allows for the flexibility of
                # using iterables, which is useful for large data structures.
                demonstrations = cast(Iterable[types.Trajectory], demonstrations)

                for traj in demonstrations:
                    for i, (obs, act, next_obs) in enumerate(
                        zip(traj.obs[:-1], traj.acts, traj.obs[1:]),
                    ):
                        flat_trans = self._preprocess_transition(obs, act, next_obs)
                        transitions.setdefault(i, []).append(flat_trans)
            elif isinstance(first_item, Mapping):
                # analogous to cast above.
                demonstrations = cast(Iterable[types.TransitionMapping], demonstrations)

                def to_np_maybe_dictobs(x):
                    if isinstance(x, types.DictObs):
                        return x
                    else:
                        return util.safe_to_numpy(x, warn=True)

                for batch in demonstrations:
                    obs = to_np_maybe_dictobs(batch["obs"])
                    acts = util.safe_to_numpy(batch["acts"], warn=True)
                    next_obs = to_np_maybe_dictobs(batch.get("next_obs"))
                    transitions.update(self._get_demo_from_batch(obs, acts, next_obs))
            else:
                raise TypeError(
                    f"Unsupported demonstration type {type(demonstrations)}",
                )
        else:
            raise TypeError(f"Unsupported demonstration type {type(demonstrations)}")

        self.transitions = {k: np.stack(v, axis=0) for k, v in transitions.items()}

        if not self.is_stationary and None in self.transitions:
            raise ValueError(
                "Non-stationary model incompatible with non-trajectory demonstrations.",
            )
        if self.is_stationary:
            self.transitions = {
                None: np.concatenate(list(self.transitions.values()), axis=0),
            }

    def train(self) -> None:
        """Fits the density model to demonstration data `self.transitions`."""
        # if requested, we'll scale demonstration transitions so that they have
        # zero mean and unit variance (i.e. all components are equally important)
        self._scaler = preprocessing.StandardScaler(
            with_mean=self.standardise,
            with_std=self.standardise,
        )
        flattened_dataset = np.concatenate(list(self.transitions.values()), axis=0)
        self._scaler.fit(flattened_dataset)

        # now fit density model
        self._density_models = {
            k: self._fit_density(self._scaler.transform(v))
            for k, v in self.transitions.items()
        }

    def _fit_density(self, transitions: np.ndarray) -> neighbors.KernelDensity:
        density_model = neighbors.KernelDensity(
            kernel=self.kernel,
            bandwidth=self.kernel_bandwidth,
        )
        density_model.fit(transitions)
        return density_model

    def _preprocess_transition(
        self,
        obs: types.Observation,
        act: np.ndarray,
        next_obs: Optional[types.Observation],
    ) -> np.ndarray:
        """Compute flattened transition on subset specified by `self.density_type`."""
        flattened_obs = space_utils.flatten(
            self.venv.observation_space,
            types.maybe_unwrap_dictobs(obs),
        )
        flattened_obs = _check_data_is_np_array(flattened_obs, "observation")
        if self.density_type == DensityType.STATE_DENSITY:
            return flattened_obs
        elif self.density_type == DensityType.STATE_ACTION_DENSITY:
            flattened_action = space_utils.flatten(self.venv.action_space, act)
            flattened_action = _check_data_is_np_array(flattened_action, "action")
            return np.concatenate([flattened_obs, flattened_action])
        elif self.density_type == DensityType.STATE_STATE_DENSITY:
            assert next_obs is not None
            flat_next_obs = space_utils.flatten(
                self.venv.observation_space,
                types.maybe_unwrap_dictobs(next_obs),
            )
            flat_next_obs = _check_data_is_np_array(flat_next_obs, "observation")
            assert type(flattened_obs) is type(flat_next_obs)

            return np.concatenate([flattened_obs, flat_next_obs])
        else:
            raise ValueError(f"Unknown density type {self.density_type}")

    def __call__(
        self,
        state: types.Observation,
        action: np.ndarray,
        next_state: types.Observation,
        done: np.ndarray,
        steps: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""Compute reward from given (s,a,s') transition batch.

        This handles *batches* of observations, since it's designed to work with
        VecEnvs.

        Args:
            state: current batch of observations.
            action: batch of actions that agent took in response to those
                observations.
            next_state: batch of observations encountered after the
                agent took those actions.
            done: is it terminal state?
            steps: What timestep is this from? Used if `self.is_stationary` is false,
                otherwise ignored.

        Returns:
            Array of scalar rewards of the form `r_t(s,a,s') = \log \hat p_t(s,a,s')`
            (one for each environment), where `\log \hat p` is the underlying density
            model (and may be independent of s', a, or t, depending on options passed
            to constructor).

        Raises:
            ValueError: Non-stationary model (`self.is_stationary` false) and `steps`
                is not provided.
        """
        if not self.is_stationary and steps is None:
            raise ValueError("steps must be provided with non-stationary models")

        del done  # TODO(adam): should we handle terminal state specially in any way?

        rew_list = []
        assert len(state) == len(action) and len(state) == len(next_state)
        state = types.maybe_wrap_in_dictobs(state)
        next_state = types.maybe_wrap_in_dictobs(next_state)
        for idx, (obs, act, next_obs) in enumerate(zip(state, action, next_state)):
            flat_trans = self._preprocess_transition(obs, act, next_obs)
            assert self._scaler is not None
            scaled_padded_trans = self._scaler.transform(flat_trans[np.newaxis])
            if self.is_stationary:
                rew = self._density_models[None].score(scaled_padded_trans)
            else:
                assert steps is not None
                time = steps[idx]
                if time >= len(self._density_models):
                    # Can't do anything sensible here yet. Correct solution is to use
                    # hierarchical model in which we first check whether state is
                    # absorbing, then assign either constant score or a score based on
                    # density.
                    raise ValueError(
                        f"Time {time} out of range (0, {len(self._density_models)}], "
                        "and absorbing states not currently supported",
                    )
                else:
                    time_model = self._density_models[time]
                    rew = time_model.score(scaled_padded_trans)
            rew_list.append(rew)
        rew_array = np.asarray(rew_list, dtype="float32")
        return rew_array

    def train_policy(self, n_timesteps: int = int(1e6), **kwargs: Any) -> None:
        """Train the imitation policy for a given number of timesteps.

        Args:
            n_timesteps: number of timesteps to train the policy for.
            kwargs (dict): extra arguments that will be passed to the `learn()`
                method of the imitation RL model. Refer to Stable Baselines docs for
                details.
        """
        assert self.rl_algo is not None
        self.rl_algo.set_env(self.venv_wrapped)
        self.rl_algo.learn(
            n_timesteps,
            # ensure we can see total steps for all
            # learn() calls, not just for this call
            reset_num_timesteps=False,
            callback=self.wrapper_callback,
            **kwargs,
        )
        trajs, ep_lens = self.buffering_wrapper.pop_trajectories()
        self._check_fixed_horizon(ep_lens)

    def test_policy(self, *, n_trajectories: int = 10, true_reward: bool = True):
        """Test current imitation policy on environment & give some rollout stats.

        Args:
            n_trajectories: number of rolled-out trajectories.
            true_reward: should this use ground truth reward from underlying
                environment (True), or imitation reward (False)?

        Returns:
            dict: rollout statistics collected by
                `imitation.utils.rollout.rollout_stats()`.
        """
        trajs = rollout.generate_trajectories(
            self.rl_algo,
            self.venv if true_reward else self.venv_wrapped,
            sample_until=rollout.make_min_episodes(n_trajectories),
            rng=self.rng,
        )
        # We collect `trajs` above so disregard return value from `pop_trajectories`,
        # but still call it to clear out any saved trajectories.
        self.buffering_wrapper.pop_trajectories()
        self._check_fixed_horizon((len(traj) for traj in trajs))
        reward_stats = rollout.rollout_stats(trajs)
        return reward_stats

    @property
    def policy(self) -> base_class.BasePolicy:
        assert self.rl_algo is not None
        assert self.rl_algo.policy is not None
        return self.rl_algo.policy


def _check_data_is_np_array(data: space_utils.FlatType, name: str) -> np.ndarray:
    """Raises error if the flattened data is not a numpy array."""
    assert isinstance(data, np.ndarray), (
        "The density estimator only supports spaces that "
        f"flatten to a numpy array but the {name} space "
        f"flattens to {type(data)}",
    )
    return data



"""Miscellaneous utility methods."""

import datetime
import functools
import itertools
import os
import pathlib
import uuid
import warnings
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
)

import gymnasium as gym
import numpy as np
import torch as th
from stable_baselines3.common import monitor, policies
from stable_baselines3.common.vec_env import DummyVecEnv, SubprocVecEnv, VecEnv

from imitation.data.types import AnyPath


def save_policy(policy: policies.BasePolicy, policy_path: AnyPath) -> None:
    """Save policy to a path.

    Args:
        policy: policy to save.
        policy_path: path to save policy to.
    """
    th.save(policy, parse_path(policy_path))


def oric(x: np.ndarray) -> np.ndarray:
    """Optimal rounding under integer constraints.

    Given a vector of real numbers such that the sum is an integer, returns a vector
    of rounded integers that preserves the sum and which minimizes the Lp-norm of the
    difference between the rounded and original vectors for all p >= 1. Algorithm from
    https://arxiv.org/abs/1501.00014. Runs in O(n log n) time.

    Args:
        x: A 1D vector of real numbers that sum to an integer.

    Returns:
        A 1D vector of rounded integers, preserving the sum.
    """
    rounded = np.floor(x)
    shortfall = x - rounded

    # The total shortfall should be *exactly* an integer, but we
    # round to account for numerical error.
    total_shortfall = np.round(shortfall.sum()).astype(int)
    indices = np.argsort(-shortfall)

    # Apportion the total shortfall to the elements in order of
    # decreasing shortfall.
    rounded[indices[:total_shortfall]] += 1
    return rounded.astype(int)


def make_unique_timestamp() -> str:
    """Timestamp, with random uuid added to avoid collisions."""
    ISO_TIMESTAMP = "%Y%m%d_%H%M%S"
    timestamp = datetime.datetime.now().strftime(ISO_TIMESTAMP)
    random_uuid = uuid.uuid4().hex[:6]
    return f"{timestamp}_{random_uuid}"


def make_vec_env(
    env_name: str,
    *,
    rng: np.random.Generator,
    n_envs: int = 8,
    parallel: bool = False,
    log_dir: Optional[str] = None,
    max_episode_steps: Optional[int] = None,
    post_wrappers: Optional[Sequence[Callable[[gym.Env, int], gym.Env]]] = None,
    env_make_kwargs: Optional[Mapping[str, Any]] = None,
) -> VecEnv:
    """Makes a vectorized environment.

    Args:
        env_name: The Env's string id in Gym.
        rng: The random state to use to seed the environment.
        n_envs: The number of duplicate environments.
        parallel: If True, uses SubprocVecEnv; otherwise, DummyVecEnv.
        log_dir: If specified, saves Monitor output to this directory.
        max_episode_steps: If specified, wraps each env in a TimeLimit wrapper
            with this episode length. If not specified and `max_episode_steps`
            exists for this `env_name` in the Gym registry, uses the registry
            `max_episode_steps` for every TimeLimit wrapper (this automatic
            wrapper is the default behavior when calling `gym.make`). Otherwise
            the environments are passed into the VecEnv unwrapped.
        post_wrappers: If specified, iteratively wraps each environment with each
            of the wrappers specified in the sequence. The argument should be a Callable
            accepting two arguments, the Env to be wrapped and the environment index,
            and returning the wrapped Env.
        env_make_kwargs: The kwargs passed to `spec.make`.

    Returns:
        A VecEnv initialized with `n_envs` environments.
    """
    # Resolve the spec outside of the subprocess first, so that it is available to
    # subprocesses running `make_env` via automatic pickling.
    # Just to ensure packages are imported and spec is properly resolved
    tmp_env = gym.make(env_name)
    tmp_env.close()
    spec = tmp_env.spec
    env_make_kwargs = env_make_kwargs or {}

    def make_env(i: int, this_seed: int) -> gym.Env:
        # Previously, we directly called `gym.make(env_name)`, but running
        # `imitation.scripts.train_adversarial` within `imitation.scripts.parallel`
        # created a weird interaction between Gym and Ray -- `gym.make` would fail
        # inside this function for any of our custom environment unless those
        # environments were also `gym.register()`ed inside `make_env`. Even
        # registering the custom environment in the scope of `make_vec_env` didn't
        # work. For more discussion and hypotheses on this issue see PR #160:
        # https://github.com/HumanCompatibleAI/imitation/pull/160.
        assert env_make_kwargs is not None  # Note: to satisfy mypy
        assert spec is not None  # Note: to satisfy mypy
        env = gym.make(spec, max_episode_steps=max_episode_steps, **env_make_kwargs)

        # Seed each environment with a different, non-sequential seed for diversity
        # (even if caller is passing us sequentially-assigned base seeds). int() is
        # necessary to work around gym bug where it chokes on numpy int64s.
        env.reset(seed=int(this_seed))
        # NOTE: we do it here rather than on the final VecEnv, because
        # that would set the same seed for all the environments.

        # Use Monitor to record statistics needed for Baselines algorithms logging
        # Optionally, save to disk
        log_path = None
        if log_dir is not None:
            log_subdir = os.path.join(log_dir, "monitor")
            os.makedirs(log_subdir, exist_ok=True)
            log_path = os.path.join(log_subdir, f"mon{i:03d}")

        env = monitor.Monitor(env, log_path)

        if post_wrappers:
            for wrapper in post_wrappers:
                env = wrapper(env, i)

        return env

    env_seeds = make_seeds(rng, n_envs)
    env_fns: List[Callable[[], gym.Env]] = [
        functools.partial(make_env, i, s) for i, s in enumerate(env_seeds)
    ]
    if parallel:
        # See GH hill-a/stable-baselines issue #217
        return SubprocVecEnv(env_fns, start_method="forkserver")
    else:
        return DummyVecEnv(env_fns)


@overload
def make_seeds(
    rng: np.random.Generator,
) -> int:
    ...


@overload
def make_seeds(rng: np.random.Generator, n: int) -> List[int]:
    ...


def make_seeds(
    rng: np.random.Generator,
    n: Optional[int] = None,
) -> Union[Sequence[int], int]:
    """Generate n random seeds from a random state.

    Args:
        rng: The random state to use to generate seeds.
        n: The number of seeds to generate.

    Returns:
        A list of n random seeds.
    """
    seeds_arr = rng.integers(0, (1 << 31) - 1, (n if n is not None else 1,))
    seeds: List[int] = seeds_arr.tolist()
    if n is None:
        return seeds[0]
    else:
        return seeds


def docstring_parameter(*args, **kwargs):
    """Treats the docstring as a format string, substituting in the arguments."""

    def helper(obj):
        obj.__doc__ = obj.__doc__.format(*args, **kwargs)
        return obj

    return helper


T = TypeVar("T")


def endless_iter(iterable: Iterable[T]) -> Iterator[T]:
    """Generator that endlessly yields elements from `iterable`.

    >>> x = range(2)
    >>> it = endless_iter(x)
    >>> next(it)
    0
    >>> next(it)
    1
    >>> next(it)
    0

    Args:
        iterable: The non-iterator iterable object to endlessly iterate over.

    Returns:
        An iterator that repeats the elements in `iterable` forever.

    Raises:
        ValueError: if iterable is an iterator -- that will be exhausted, so
            cannot be iterated over endlessly.
    """
    if iter(iterable) == iterable:
        raise ValueError("endless_iter needs a non-iterator Iterable.")

    _, iterable = get_first_iter_element(iterable)
    return itertools.chain.from_iterable(itertools.repeat(iterable))


def safe_to_tensor(array: Union[np.ndarray, th.Tensor], **kwargs) -> th.Tensor:
    """Converts a NumPy array to a PyTorch tensor.

    The data is copied in the case where the array is non-writable. Unfortunately if
    you just use `th.as_tensor` for this, an ugly warning is logged and there's
    undefined behavior if you try to write to the tensor.

    Args:
        array: The array to convert to a PyTorch tensor.
        kwargs: Additional keyword arguments to pass to `th.as_tensor`.

    Returns:
        A PyTorch tensor with the same content as `array`.
    """
    if isinstance(array, np.ndarray) and not array.flags.writeable:
        array = array.copy()

    return th.as_tensor(array, **kwargs)


@overload
def safe_to_numpy(obj: Union[np.ndarray, th.Tensor], warn: bool = False) -> np.ndarray:
    ...


@overload
def safe_to_numpy(obj: None, warn: bool = False) -> None:
    ...


def safe_to_numpy(
    obj: Optional[Union[np.ndarray, th.Tensor]],
    warn: bool = False,
) -> Optional[np.ndarray]:
    """Convert torch tensor to numpy.

    If the object is already a numpy array, return it as is.
    If the object is none, returns none.

    Args:
        obj: torch tensor object to convert to numpy array
        warn: if True, warn if the object is not already a numpy array. Useful for
            warning the user of a potential performance hit if a torch tensor is
            not the expected input type.

    Returns:
        Object converted to numpy array
    """
    if obj is None:
        # We ignore the type due to https://github.com/google/pytype/issues/445
        return None  # pytype: disable=bad-return-type
    elif isinstance(obj, np.ndarray):
        return obj
    else:
        if warn:
            warnings.warn(
                "Converted tensor to numpy array, might affect performance. "
                "Make sure this is the intended behavior.",
            )
        return obj.detach().cpu().numpy()


def tensor_iter_norm(
    tensor_iter: Iterable[th.Tensor],
    ord: Union[int, float] = 2,  # noqa: A002
) -> th.Tensor:
    """Compute the norm of a big vector that is produced one tensor chunk at a time.

    Args:
        tensor_iter: an iterable that yields tensors.
        ord: order of the p-norm (can be any int or float except 0 and NaN).

    Returns:
        Norm of the concatenated tensors.

    Raises:
        ValueError: ord is 0 (unsupported).
    """
    if ord == 0:
        raise ValueError("This function cannot compute p-norms for p=0.")
    norms = []
    for tensor in tensor_iter:
        norms.append(th.norm(tensor.flatten(), p=ord))
    norm_tensor = th.as_tensor(norms)
    # Norm of the norms is equal to the norm of the concatenated tensor.
    # th.norm(norm_tensor) = sum(norm**ord for norm in norm_tensor)**(1/ord)
    # = sum(sum(x**ord for x in tensor) for tensor in tensor_iter)**(1/ord)
    # = sum(x**ord for x in tensor for tensor in tensor_iter)**(1/ord)
    # = th.norm(concatenated tensors)
    return th.norm(norm_tensor, p=ord)


def get_first_iter_element(iterable: Iterable[T]) -> Tuple[T, Iterable[T]]:
    """Get first element of an iterable and a new fresh iterable.

    The fresh iterable has the first element added back using ``itertools.chain``.
    If the iterable is not an iterator, this is equivalent to
    ``(next(iter(iterable)), iterable)``.

    Args:
        iterable: The iterable to get the first element of.

    Returns:
        A tuple containing the first element of the iterable, and a fresh iterable
        with all the elements.

    Raises:
        ValueError: `iterable` is empty -- the first call to it returns no elements.
    """
    iterator = iter(iterable)
    try:
        first_element = next(iterator)
    except StopIteration:
        raise ValueError(f"iterable {iterable} had no elements to iterate over.")

    return_iterable: Iterable[T]
    if iterator == iterable:
        # `iterable` was an iterator. Getting `first_element` will have removed it
        # from `iterator`, so we need to add a fresh iterable with `first_element`
        # added back in.
        return_iterable = itertools.chain([first_element], iterator)
    else:
        # `iterable` was not an iterator; we can just return `iterable`.
        # `iter(iterable)` will give a fresh iterator containing the first element.
        # It's preferable to return `iterable` without modification so that users
        # can generate new iterators from it as needed.
        return_iterable = iterable

    return first_element, return_iterable


def parse_path(
    path: AnyPath,
    allow_relative: bool = True,
    base_directory: Optional[pathlib.Path] = None,
) -> pathlib.Path:
    """Parse a path to a `pathlib.Path` object.

    All resulting paths are resolved, absolute paths. If `allow_relative` is True,
    then relative paths are allowed as input, and are resolved relative to the
    current working directory, or relative to `base_directory` if it is
    specified.

    Args:
        path: The path to parse. Can be a string, bytes, or `os.PathLike`.
        allow_relative: If True, then relative paths are allowed as input, and
            are resolved relative to the current working directory. If False,
            an error is raised if the path is not absolute.
        base_directory: If specified, then relative paths are resolved relative
            to this directory, instead of the current working directory.

    Returns:
        A `pathlib.Path` object.

    Raises:
        ValueError: If `allow_relative` is False and the path is not absolute.
        ValueError: If `base_directory` is specified and `allow_relative` is
            False.
    """
    if base_directory is not None and not allow_relative:
        raise ValueError(
            "If `base_directory` is specified, then `allow_relative` must be True.",
        )

    parsed_path: pathlib.Path
    if isinstance(path, pathlib.Path):
        parsed_path = path
    elif isinstance(path, str):
        parsed_path = pathlib.Path(path)
    elif isinstance(path, bytes):
        parsed_path = pathlib.Path(path.decode())
    else:
        parsed_path = pathlib.Path(str(path))

    if parsed_path.is_absolute():
        return parsed_path
    else:
        if allow_relative:
            base_directory = base_directory or pathlib.Path.cwd()
            # relative to current working directory
            return base_directory / parsed_path
        else:
            raise ValueError(f"Path {str(parsed_path)} is not absolute")


def parse_optional_path(
    path: Optional[AnyPath],
    allow_relative: bool = True,
    base_directory: Optional[pathlib.Path] = None,
) -> Optional[pathlib.Path]:
    """Parse an optional path to a `pathlib.Path` object.

    All resulting paths are resolved, absolute paths. If `allow_relative` is True,
    then relative paths are allowed as input, and are resolved relative to the
    current working directory, or relative to `base_directory` if it is
    specified.

    Args:
        path: The path to parse. Can be a string, bytes, or `os.PathLike`.
        allow_relative: If True, then relative paths are allowed as input, and
            are resolved relative to the current working directory. If False,
            an error is raised if the path is not absolute.
        base_directory: If specified, then relative paths are resolved relative
            to this directory, instead of the current working directory.

    Returns:
        A `pathlib.Path` object, or None if `path` is None.
    """
    if path is None:
        return None
    else:
        return parse_path(path, allow_relative, base_directory)


def split_in_half(x: int) -> Tuple[int, int]:
    """Split an integer in half, rounding up.

    This is to ensure that the two halves sum to the original integer.

    Args:
        x: The integer to split.

    Returns:
        A tuple containing the two halves of `x`.
    """
    half = x // 2
    return half, x - half


def clear_screen() -> None:
    """Clears the console screen."""
    if os.name == "nt":  # Windows
        os.system("cls")  # pragma: no cover
    else:
        os.system("clear")



"""Tests `imitation.algorithms.sqil`."""
from typing import Any, Dict, Optional, Type
from unittest import mock

import numpy as np
import pytest
from stable_baselines3 import ddpg, dqn, sac, td3
from stable_baselines3.common import off_policy_algorithm, policies, vec_env
from stable_baselines3.common.evaluation import evaluate_policy

from imitation.algorithms import sqil
from imitation.testing import expert_trajectories, reward_improvement

EXPERT_DATA_TYPES = ["trajectories", "transitions"]
RL_ALGOS_CONT_ACTIONS = [ddpg.DDPG, sac.SAC, td3.TD3]


def get_demos(
    env_name: str,
    rng: np.random.Generator,
    pytestconfig: pytest.Config,
    data_type: str,
):
    cache = pytestconfig.cache
    assert cache is not None
    return expert_trajectories.make_expert_transition_loader(
        cache_dir=cache.mkdir("experts"),
        batch_size=4,
        expert_data_type=data_type,
        env_name=env_name,
        rng=rng,
        num_trajectories=60,
        shuffle=False,
    )


@pytest.mark.parametrize("expert_data_type", EXPERT_DATA_TYPES)
def test_sqil_demonstration_buffer(
    rng: np.random.Generator,
    pytestconfig: pytest.Config,
    cartpole_venv: vec_env.VecEnv,
    expert_data_type: str,
):
    policy = "MlpPolicy"
    model = sqil.SQIL(
        venv=cartpole_venv,
        demonstrations=get_demos(
            "seals/CartPole-v0",
            rng,
            pytestconfig,
            expert_data_type,
        ),
        policy=policy,
    )

    assert isinstance(model.policy, policies.BasePolicy)
    assert isinstance(model.rl_algo.replay_buffer, sqil.SQILReplayBuffer)
    expert_buffer = model.rl_algo.replay_buffer.expert_buffer

    # Check that demonstrations are stored in the replay buffer correctly
    demonstrations = get_demos("seals/CartPole-v0", rng, pytestconfig, "transitions")
    n_samples = len(demonstrations)
    assert len(model.rl_algo.replay_buffer.expert_buffer.observations) == n_samples
    for i in range(n_samples):
        obs = expert_buffer.observations[i]
        act = expert_buffer.actions[i]
        assert expert_buffer.next_observations is not None
        next_obs = expert_buffer.next_observations[i]
        done = expert_buffer.dones[i]

        np.testing.assert_array_equal(obs[0], demonstrations.obs[i])
        np.testing.assert_array_equal(act[0], demonstrations.acts[i])
        np.testing.assert_array_equal(next_obs[0], demonstrations.next_obs[i])
        np.testing.assert_array_equal(done, demonstrations.dones[i])


def _test_sqil_no_crash(
    rng: np.random.Generator,
    pytestconfig: pytest.Config,
    venv: vec_env.VecEnv,
    env_name: str,
    rl_algo_class: Type[off_policy_algorithm.OffPolicyAlgorithm],
    rl_kwargs: Optional[Dict[str, Any]] = None,
):
    policy = "MlpPolicy"
    model = sqil.SQIL(
        venv=venv,
        demonstrations=get_demos(env_name, rng, pytestconfig, "transitions"),
        policy=policy,
        rl_algo_class=rl_algo_class,
        rl_kwargs=rl_kwargs,
    )
    model.train(total_timesteps=500)


def test_sqil_no_crash_discrete(
    rng: np.random.Generator,
    pytestconfig: pytest.Config,
    cartpole_venv: vec_env.VecEnv,
):
    _test_sqil_no_crash(
        rng,
        pytestconfig,
        cartpole_venv,
        "seals/CartPole-v0",
        rl_algo_class=dqn.DQN,
        rl_kwargs=dict(learning_starts=100),
    )


@pytest.mark.parametrize("rl_algo_class", RL_ALGOS_CONT_ACTIONS)
def test_sqil_no_crash_continuous(
    rng: np.random.Generator,
    pytestconfig: pytest.Config,
    pendulum_single_venv: vec_env.VecEnv,
    rl_algo_class: Type[off_policy_algorithm.OffPolicyAlgorithm],
):
    _test_sqil_no_crash(
        rng,
        pytestconfig,
        pendulum_single_venv,
        "Pendulum-v1",
        rl_algo_class=rl_algo_class,
    )


def _test_sqil_few_demonstrations(
    rng: np.random.Generator,
    pytestconfig: pytest.Config,
    venv: vec_env.VecEnv,
    env_name: str,
    rl_algo_class: Type[off_policy_algorithm.OffPolicyAlgorithm],
    rl_kwargs: Optional[Dict[str, Any]] = None,
):
    demonstrations = get_demos(env_name, rng, pytestconfig, "transitions")
    demonstrations = demonstrations[:5]

    policy = "MlpPolicy"
    model = sqil.SQIL(
        venv=venv,
        demonstrations=demonstrations,
        policy=policy,
        rl_algo_class=rl_algo_class,
        rl_kwargs=rl_kwargs,
    )
    model.train(total_timesteps=1_00)


def test_sqil_few_demonstrations_discrete(
    rng: np.random.Generator,
    pytestconfig: pytest.Config,
    cartpole_venv: vec_env.VecEnv,
):
    _test_sqil_few_demonstrations(
        rng,
        pytestconfig,
        cartpole_venv,
        "seals/CartPole-v0",
        rl_algo_class=dqn.DQN,
        rl_kwargs=dict(learning_starts=10, seed=42),
    )


@pytest.mark.parametrize("rl_algo_class", RL_ALGOS_CONT_ACTIONS)
def test_sqil_few_demonstrations_continuous(
    rng: np.random.Generator,
    pytestconfig: pytest.Config,
    pendulum_single_venv: vec_env.VecEnv,
    rl_algo_class: Type[off_policy_algorithm.OffPolicyAlgorithm],
):
    _test_sqil_few_demonstrations(
        rng,
        pytestconfig,
        pendulum_single_venv,
        "Pendulum-v1",
        rl_algo_class=rl_algo_class,
        rl_kwargs=dict(seed=42),
    )


def _test_sqil_performance(
    rng: np.random.Generator,
    pytestconfig: pytest.Config,
    venv: vec_env.VecEnv,
    env_name: str,
    rl_algo_class: Type[off_policy_algorithm.OffPolicyAlgorithm],
    rl_kwargs: Optional[Dict[str, Any]] = None,
):
    SEED = 42
    demonstrations = get_demos(env_name, rng, pytestconfig, "transitions")
    model = sqil.SQIL(
        venv=venv,
        demonstrations=demonstrations,
        policy="MlpPolicy",
        rl_algo_class=rl_algo_class,
        rl_kwargs=rl_kwargs,
    )

    venv.seed(SEED)
    rewards_before, _ = evaluate_policy(
        model.policy,
        venv,
        100,
        return_episode_rewards=True,
    )

    model.train(total_timesteps=1_000)

    venv.seed(SEED)
    rewards_after, _ = evaluate_policy(
        model.policy,
        venv,
        100,
        return_episode_rewards=True,
    )

    assert reward_improvement.is_significant_reward_improvement(
        rewards_before,  # type:ignore[arg-type]
        rewards_after,  # type:ignore[arg-type]
    )


def test_sqil_performance_discrete(
    rng: np.random.Generator,
    pytestconfig: pytest.Config,
    cartpole_venv: vec_env.VecEnv,
):
    _test_sqil_performance(
        rng,
        pytestconfig,
        cartpole_venv,
        "seals/CartPole-v0",
        rl_algo_class=dqn.DQN,
        rl_kwargs=dict(
            learning_starts=500,
            learning_rate=0.002,
            batch_size=220,
            seed=42,
        ),
    )


@pytest.mark.skip(reason="This test is flaky.")
@pytest.mark.parametrize("rl_algo_class", RL_ALGOS_CONT_ACTIONS)
def test_sqil_performance_continuous(
    rng: np.random.Generator,
    pytestconfig: pytest.Config,
    pendulum_single_venv: vec_env.VecEnv,
    rl_algo_class: Type[off_policy_algorithm.OffPolicyAlgorithm],
):  # pragma: no cover
    rl_kwargs = dict(
        learning_starts=500,
        learning_rate=0.001,
        gamma=0.95,
        seed=42,
    )
    if rl_algo_class == ddpg.DDPG:
        rl_kwargs["gamma"] = 0.99
        rl_kwargs["learning_starts"] = 100
    _test_sqil_performance(
        rng,
        pytestconfig,
        pendulum_single_venv,
        "Pendulum-v1",
        rl_algo_class=rl_algo_class,
        rl_kwargs=rl_kwargs,
    )


@pytest.mark.parametrize("illegal_kw", ["replay_buffer_class", "replay_buffer_kwargs"])
def test_sqil_constructor_raises(illegal_kw: str):
    with pytest.raises(ValueError, match=".*SQIL uses a custom replay buffer.*"):
        sqil.SQIL(
            venv=mock.MagicMock(spec=vec_env.VecEnv),
            demonstrations=None,
            policy="MlpPolicy",
            rl_kwargs={illegal_kw: None},
        )
